from typing import Union




class PyGraphVisual:
    """
    Facilitate data preparation for graph visualization that uses the Cytoscape.js library

        NOTE: this is a fork of a class by the same name in the sister project BrainAnnex.org (where the leading
          development for this module takes place)
    """


    def __init__(self, db=None):
        self.db = db                    # Object of "GraphAccess" class

        self.nodes = []                 # The node data needed by the Cytoscape graph visualization.
                                        #   A list of dicts defining nodes.
                                        #   NODES must contain the keys 'id' and '_node_labels'
                                        #   EXAMPLE (2 nodes):
                                        #   [{'id': 1, '_node_labels': ['PERSON'], 'name': 'Julian'},
                                        #    {'id': 2, '_node_labels': ['CAR'], 'color': 'white'}]

        self.edges = []                 # The edge data needed by the Cytoscape graph visualization.
                                        #   A list of dicts defining edges.
                                        #   EDGES must contain the keys 'name', 'source', and 'target'
                                        #       (and presumably 'id' is required as well?)
                                        # EXAMPLE (1 edge):
                                        #   [{'name': 'OWNS', 'source': 1, 'target': 2, 'id': 'edge-1'}]
                                        #   The above assumes that nodes 1 and 2 exist

        self.color_mapping = {}         # Mapping a node label to its interior color (the edge color is an automatic variation)
                                        # EXAMPLE:  {'PERSON': 'cyan', 'CAR': 'orange'}

        self.caption_mapping = {}       # Mapping a node label to its caption (field name to use on the graph)
                                        # EXAMPLE:  {'PERSON': 'name', 'CAR': 'color'}

        self._all_node_ids = []         # List of all the node id's added to the graph so far;
                                        #   used for optional prevention of duplicates

        self._next_available_edge_id = 1 # An auto-increment value


        self.EXTRA_COLORS =  {          # Convenient extra colors, not available thru standard CSS names
            "graph_green": '#8DCC92',
            "graph_darkgreen": '#56947E',
            "graph_teal": '#569481',
            "graph_orange": '#F79768',
            "graph_blue": '#4C8EDB',
            "graph_red": '#F16668',
            "graph_darkbrown": '#604A0D',
            "graph_lightbrown": '#D9C8AD',
            "graph_orchid": '#C990C1',
            "graph_gold": '#FFC455'     # Has a bit more orange than 'gold'
        }



    def __str__(self) -> str:
        """
        Return an overview description of this object

        :return:    A string with a description of this object
        """
        s = f"Object describing a graph containing {len(self.nodes)} nodes " \
            f"and {len(self.edges)} edges:\n"

        s += f"    Graph Node structure: {self.nodes}\n"
        s += f"    Graph Edge structure: {self.edges}\n"
        s += f"    Graph color mapping: {self.color_mapping}\n"
        s += f"    Graph caption mapping: {self.caption_mapping}"

        return s



    def get_graph_data(self) -> dict:
        """
        Return a data dictionary with all the relevant visualization info
        (typically, to pass to the front-end, for eventual use by the Cytoscape.js library)

        :return:    A dict with 4 keys - "nodes", "edges", "color_mapping" and "caption_mapping"
                        For more details, see object constructor
        """
        #TODO: perhaps add a 2nd arg, to specify just one element to be returned
        return {"nodes": self.nodes,
                "edges": self.edges,
                "color_mapping": self.color_mapping,
                "caption_mapping": self.caption_mapping}



    def add_node(self, node_id :int|str, labels="", properties=None) -> None:
        """
        From the given data about a database node, assemble and cumulatively store the data
        in a format expected by Cytoscape.js (the visualization front-end).

        Calls with a duplicate node_id are silently disregarded.

        NO database operation is performed.  You need to first extract, or have on hand, that node data.

        EXAMPLE:    given a call to:  add_node(node_id=1, labels=['PERSON'], data={'name': 'Julian'})
                    then the internal format, cumulatively added to self.nodes, will be:
                    {'id': 1, '_node_labels': ['PERSON'], 'name': 'Julian'}

        :param node_id:     Either an integer or a string to uniquely identify this node in the graph;
                                it will be used to specify the start/end nodes of edges.
                                Typically, use either the internal database ID, or some URI.
                                Records with a duplicate node_id are silently disregarded
        :param labels:      A string, or list of strings, with the node's label(s) used in the graph database;
                                if not used, pass an empty string
        :param properties:  A dict with all the node properties of interest
        :return:            None
        """
        assert isinstance(node_id, (int, str)), \
            "add_node(): the argument `node_id` must be an integer or a string"

        if type(node_id) == str:
            assert node_id != "", "add_node(): cannot use an empty string for the argument `node_id`"


        if node_id in self._all_node_ids:
            return      # Silently disregard duplicates.   TODO: use a "sorted list" for `self._all_node_ids`

        if properties is None:
            properties = {}

        d = properties.copy()     # Make an independent clone of the data dict


        d["id"] = node_id

        if type(labels) == str:
            labels = [labels]   # Turn into a list, if passed as a single string value

        d["_node_labels"] = labels

        self.nodes.append(d)
        self._all_node_ids.append(node_id)



    def add_edge(self, from_node :str|int, to_node :str|int,
                 name :str, edge_id=None, properties=None) -> None:
        """
        Prepare and store the data for 1 edge, in a format expected by the visualization front-end.

        EXAMPLE:    given a call to:  add_edge(from_node=1, to_node=2, name='OWNS', edge_id='edge-1')
                    then the internal format, cumulatively added to self.edges, will be:
                    {'name': 'OWNS', 'source': 1, 'target': 2, 'id': 'edge-1'}

        Note that the id's, in this context, is whatever we pass to the visualization module for the nodes;
        not necessarily related to the internal database ID's

        :param from_node:   Integer or a string uniquely identify the "id" of the node where the edge originates
        :param to_node:     Integer or a string uniquely identify the "id" of the node where the edge ends
        :param name:        Name of the relationship
        :param edge_id:     [OPTIONAL]  If not provided, strings such as "edge-123" are used,
                                with auto-generated consecutive integers
        :param properties:  [OPTIONAL]  A dict with all the edge properties of interest.
                                If keys with conflicting names are present ('name', 'source', 'target', 'id'),
                                an underscore is prefixed to them

        :return:            None
        """
        # TODO: check if the edge already exists, with same name
        # TODO: unclear if edge id's are really needed by Cytoscape.js

        d = {"name": name, "source": from_node, "target": to_node}

        if edge_id is not None:
            d["id"] = edge_id
        else:
            # Use an auto-increment value of the form "edge-n" for some n
            d["id"] = f"edge-{self._next_available_edge_id}"
            self._next_available_edge_id += 1        # Maintain an auto-increment value for edge ID's

        if properties:
            if "name" in properties:
                properties["_name"] = properties["name"]
                del properties["name"]
            if "source" in properties:
                properties["_source"] = properties["source"]
                del properties["source"]
            if "target" in properties:
                properties["_target"] = properties["target"]
                del properties["target"]
            if "id" in properties:
                properties["_id"] = properties["id"]
                del properties["id"]

            d.update(properties)    # Also store the edge properties in its dict

        self.edges.append(d)



    def assign_caption(self, label :str, caption="name") -> None:
        """
        Incrementally assign and store a mapping from label name to caption (name of field to use on the display).
        Additional calls to this function will register further assignments.

        EXAMPLES:   assign_caption(label='PERSON', caption='name')
                    assign_caption(label='CAR', caption='color')

        :param label:   The name of a node label
        :param caption: The name of the node field to use on the display (by default, "name")
        :return:        None
        """
        assert type(label) == str, \
            "assign_caption(): the argument `label` must be a string"

        self.caption_mapping[label] = caption



    def assign_color_mapping(self, label :str, color :str) -> None:
        """
        Incrementally assign and store a mapping from label name to color to use for nodes having that label.
        Additional calls to this function will register further assignments.

        EXAMPLES:   assign_color_mapping(label='PERSON', color='purple')
                    assign_color_mapping(label='CAR', color='#FF0088')

        :param label:   The name of a node label
        :param color:   The name (hex string or standard CSS name) of the color to use on the display,
                            OR one of the non-standard names provided by this class,
                            in the object variable self.EXTRA_COLORS
        :return:        None
        """
        if color in self.EXTRA_COLORS:
            color = self.EXTRA_COLORS[color]     # Convert the non-standard name into a numerical value

        self.color_mapping[label] = color




    ############   The methods below require a database connection   ############


    def prepare_recordset(self, id_list :[int|str]) -> [dict]:
        """
        Given a list of node internal id's, construct and return a dataset of their properties,
        plus the special fields `internal_id` and `_node_labels`

        :param id_list: A list of internal id's of database nodes

        :return:        A list of dicts, with the node properties plus the special fields `internal_id` and `_node_labels`
                            EXAMPLE:
                            [{'_internal_id': 123, '_node_labels': ['Person'], 'name': 'Julian'}]
        """
        # TODO: move to GraphAccess
        assert type(id_list) == list, \
            f"prepare_recordset(): argument `id_list` must be a list; it is of type {type(id_list)}"

        assert self.db, \
            "prepare_recordset(): missing database handle; did you pass it when instantiating PyGraphVisual(db=...) ?"

        if id_list == []:
            return []       # No data was passed

        q = f'''
            MATCH (n)
            WHERE ID(n) IN $id_list
            RETURN n
            '''

        #self.db.debug_print_query(q, {"id_list": id_list})
        return self.db.query_extended(q, {"id_list": id_list}, flatten=True)



    def prepare_graph(self, result_dataset :[dict], cumulative=False, add_edges=True, avoid_links=None) -> [int|str]:
        """
        Given a list of dictionary data containing the properties of graph-database nodes - for example,
        as returned by GraphAccess.get_nodes() - construct, and save inside this pyhon object,
        the visualization data for them.

        Each passed dictionary entry MUST have a key named "_internal_id".
        If any key named "id" is found, it get automatically renamed "_id_original" (since "id" is used by the visualization software);
        if "_id_original" already exists, an Exception is raised.  (Copies are made; the original data object isn't affected.)
        Though not required, a key named "_node_labels" is typically present as well.

        Any date/datetime value found in the database will first be "sanitized" into a string representation of the date;
        the time portion, if present, will get dropped

        :param result_dataset:  A list of dictionary data about graph-database nodes;
                                    each dict must contain an entry with the key "_internal_id".
                                    No problem if duplicates in "_internal_id" are present;
                                    only the fist of one duplicates gets processed
        :param cumulative:      If False (default) then any previous call to this function will get ignored,
                                    and a new graph is appended
        :param add_edges:       If True, all existing edges among the displayed nodes
                                    will also be part of the visualization
        :param avoid_links:     Name or list of name of links to avoid including

        :return:                A list of values with the internal databased IDs
                                    of all the nodes added to the graph structure
        """
        assert self.db, \
            "prepare_graph(): missing database handle; did you pass it when instantiating PyGraphVisual(db=...) ?"

        assert type(result_dataset) == list, \
            f"prepare_graph(): argument `id_list` must be a list; it is of type {type(result_dataset)}"

        if len(result_dataset) == 0:
            return []       # No data was passed


        if not cumulative:
            # Reset: clear out any previous graph-structure data, and reset edge number auto-increment
            self.nodes = []
            self.edges = []
            self._all_node_ids = []
            self._next_available_edge_id = 1


        id_key_renaming = False
        node_list = []      # Running list of internal databased IDs, for nodes in `result_dataset`
        for node in result_dataset:
            internal_id = node.get("_internal_id")
            assert internal_id is not None, \
                f"prepare_graph() - the following record lacks the required `internal_id` key: {node}"

            if internal_id in node_list:
                # Ignore records already seen (duplicates in `internal_id`)
                print(f"prepare_graph(): ignoring duplicate record with `internal_id` = {internal_id}")
                continue

            node_clone = node.copy()

            if "id" in node_clone:
                assert "_id_original" not in node_clone, \
                    f"prepare_graph(): keys named `id` are routinely automatically renamed `_id_original`, " \
                    f"but the latter key also already exists!  Found in: {node_clone}"
                node_clone["_id_original"] = node_clone["id"]
                del node_clone["id"]
                id_key_renaming = True

            node_list.append(internal_id)

            if "_node_labels" in node_clone:
                labels = node["_node_labels"]
            else:
                labels = ""

            self.add_node(node_id=internal_id, labels=labels,
                          properties=self.db.sanitize_date_times(node_clone, drop_time=True))


        if add_edges:
            # Search the database for any edges among any of the nodes selected for the visualization
            exclude_clause = ""
            if avoid_links:
                if type(avoid_links) == str:
                    avoid_links = [avoid_links]
                else:
                    assert type(avoid_links) == list,  \
                        f"prepare_graph(): The argument `avoid_links`, if passed, must be a string or a list"

                exclude_clause = f"AND NOT type(r) IN {avoid_links}"


            q = f'''
                MATCH (n1)-[r]->(n2) 
                WHERE ID(n1) IN $node_list AND ID(n2) IN $node_list 
                {exclude_clause}
                RETURN DISTINCT id(n1) AS from_node, id(n2) AS to_node, 
                       type(r) AS rel_name, properties(r) AS rel_props
                '''

            #self.db.debug_print_query(q, {"node_list": node_list})
            result = self.db.query(q, {"node_list": node_list})


            # "Sanitize" the records, as needed, i.e. make them suitable for JSON serialization,
            # in anticipation of eventually passing the data to JavaScript
            for edge in result:
                #print(edge)
                edge_props = self.db.sanitize_date_times(edge["rel_props"], drop_time=True)
                self.add_edge(from_node=edge["from_node"], to_node=edge["to_node"],
                              name=edge["rel_name"], properties=edge_props)

        if id_key_renaming:
            print("prepare_graph(): keys named `id` were found in one or more of the records; "
                  "they were renamed `_id_original` to avoid conflict with internal database IDs")

        return node_list



    def assemble_graph(self, id_list :[int|str]) -> ([dict],[dict]):
        """
        Given a list of node internal id's, construct and return
        the data needed by the Cytoscape graph visualization (including all properties).

        Any date/datetime value found in the database will first be "sanitized"
        into a string representation of the date;
        the time portion, if present, will get dropped

        :param id_list: A list of internal id's of database nodes
        :return:        A pair with all the data needed by the Cytoscape graph visualization:
                            1) list of dicts defining nodes
                            2) list of dicts defining edges
        """
        recordset = self.prepare_recordset(id_list)
        self.prepare_graph(result_dataset=recordset, cumulative=False, add_edges=True)
        return (self.nodes , self.edges)



    def link_node_groups(self, group1 :[int|str], group2 :[int|str]) -> None:
        """
        Search the database for any edges from any of the nodes in the 1st group, to any node in the 2nd group.
        Any located edge will be added to the visualization data stored in this object.

        :param group1:  List of the internal databased IDs of the 1st group of database nodes
        :param group2:  List of the internal databased IDs of the 2nd group of database nodes
        :return:        None
        """
        q = '''
            MATCH (n1)-[r]->(n2) 
            WHERE ID(n1) IN $group1 AND ID(n2) IN $group2 
            RETURN DISTINCT id(n1) AS from_node, id(n2) AS to_node, type(r) AS name
            '''

        result = self.db.query(q, {"group1": group1, "group2": group2})
        for edge in result:
            #print(edge)
            self.add_edge(from_node=edge["from_node"], to_node=edge["to_node"], name=edge["name"])



##########################################################################################

class PyGraphVisual_OLD:
    """
    Facilitate data preparation for graph visualization using the Cytoscape.js library.

    NOTE: this is a fork of a class by the same name in the sister project BrainAnnex.org (where the leading
          development for this module takes place)
    """


    def __init__(self, db=None):
        self.db = db                    # Object of "GraphAccess" class

        self.structure = []             # The data that defines a graph to visualize.
                                        #   A list of dicts defining nodes, and possibly edges as well.
                                        #   NODES must contain the keys 'id' and 'labels'
                                        #   EDGES must contain the keys 'name', 'source', and 'target'
                                        #       (and presumably 'id' is required as well?)
                                        # EXAMPLE (2 nodes and an edge):
                                        #   [{'id': 1, 'labels': ['PERSON'], 'name': 'Julian'},
                                        #    {'id': 2, 'labels': ['CAR'], 'color': 'white'},
                                        #    {'name': 'OWNS', 'source': 1, 'target': 2, 'id': 'edge-1'}]

        self.color_mapping = {}         # Mapping a node label to its interior color (the edge color is an automatic variation)
                                        # EXAMPLE:  {'PERSON': 'cyan', 'CAR': 'orange'}

        self.caption_mapping = {}       # Mapping a node label to its caption (field name to use on the graph)
                                        # EXAMPLE:  {'PERSON': 'name', 'CAR': 'color'}

        self._next_available_edge_id = 1 # An auto-increment value

        self._all_node_ids = []         # List of all the node id's added to the graph so far;
                                        #   used for optional prevention of duplicates


    def serialize(self) -> dict:
        """
        Extract and return a dict of relevant data from this object

        :return:
        """
        return {"structure": self.structure,
                "color_mapping": self.color_mapping,
                "caption_mapping": self.caption_mapping}



    def __str__(self) -> str:
        """
        Return an overview description of this object

        :return:    A string with a description of this object
        """
        s = f"Object describing a graph containing a total of {len(self.structure)} nodes and edges:\n"
        s += f"    Graph structure: {self.structure}\n"
        s += f"    Graph color mapping: {self.color_mapping}\n"
        s += f"    Graph caption mapping: {self.caption_mapping}"

        return s



    def get_graph_data(self) -> dict:
        """
        Return a data dictionary with all the relevant visualization info
        (typically, to pass to the front-end, for eventual use by the Cytoscape.js library)

        :return:    A dict with 3 keys - "structure", "color_mapping" and "caption_mapping"
        """
        return {"structure": self.structure,
                "color_mapping": self.color_mapping,
                "caption_mapping": self.caption_mapping}



    def add_node(self, node_id :Union[int,str], labels="", data=None) -> None:
        """
        Prepare and store the data for 1 node, in a format expected by the visualization front-end

        EXAMPLE:    {'id': 1, 'name': 'Julian', 'labels': ['PERSON']}

        :param node_id:         Either an integer or a string to uniquely identify this node in the graph;
                                    it will be used to specify the start/end nodes of edges.
                                    Typically, use either the internal database ID, or some URI
                                    Records with a duplicate node_id are silently disregarded
        :param labels:          A string, or list of strings, with the node's label(s) used in the graph database;
                                    if not used, pass an empty string
        :param data:            A dict with all other node data not already specified in any of the other arguments
        :return:                None
        """
        assert node_id != "", \
            "add_node(): cannot use an empty string for the argument `node_id`"


        if node_id in self._all_node_ids:
            return      # Silently disregard duplicates

        if data is None:
            data = {}

        d = data.copy()     # Make an independent clone of the data dict


        d["id"] = node_id

        if type(labels) == str:
            labels = [labels]

        d["labels"] = labels

        self.structure.append(d)
        self._all_node_ids.append(node_id)



    def add_edge(self, from_node :Union[str, int], to_node :Union[str, int],
                 name :str, edge_id=None, data=None) -> None:
        """
        Prepare and store the data for 1 edge, in a format expected by the visualization front-end.

        EXAMPLE:   {'name': 'OWNS', 'source': 1, 'target': 2, 'id': 'edge-1'}

        Note that 'id', in this context, is whatever we pass to the visualization module for the nodes;
        not necessarily related to the internal database ID's

        :param from_node:   Integer or a string uniquely identify the "id" of the node where the edge originates
        :param to_node:     Integer or a string uniquely identify the "id" of the node where the edge ends
        :param name:        Name of the relationship
        :param edge_id:     (OPTIONAL)  If not provided, strings such as "edge-123" are used,
                                        with auto-generated consecutive integers
                                TODO: unclear if edge id's are really needed
        :param data:        A dict with all other node data not already specified in any of the other arguments

        :return:            None
        """
        from_id = from_node
        to_id = to_node

        d = {"name": name, "source": from_id, "target": to_id}

        if edge_id is not None:
            d["id"] = edge_id
        else:
            d["id"] = f"edge-{self._next_available_edge_id}"
            self._next_available_edge_id += 1        # Maintain an auto-increment value

        if data:
            d.update(data)

        self.structure.append(d)



    def assign_caption(self, label :str, caption="name") -> None:
        """
        Assign and store a mapping from label name to caption (name of field to use on the display)

        EXAMPLES:   assign_caption(label='PERSON', caption='name')
                    assign_caption(label='CAR', caption='color')

        :param label:   The name of a node label
        :param caption: The name of the node field to use on the display (by default, "name")
        :return:        None
        """
        assert type(label) == str, \
            "assign_caption(): the argument `label` must be a string"

        self.caption_mapping[label] = caption



    def assign_color_mapping(self, label :str, color :str) -> None:
        """
        Assign and store a mapping from label name to color to use for nodes having that label

        :param label:   The name of a node label
        :param color:   The name (hex string or standard CSS name) of the color to use on the display
        :return:        None
        """
        extra_colors =  {       # Convenient extra colors, not available thru standard CSS names
            "graph_green": '#8DCC92',
            "graph_darkgreen": '#56947E',
            "graph_teal": '#569481',
            "graph_orange": '#F79768',
            "graph_blue": '#4C8EDB',
            "graph_red": '#F16668',
            "graph_darkbrown": '#604A0D',
            "graph_lightbrown": '#D9C8AD',
            "graph_orchid": '#C990C1',
            "graph_gold": '#FFC455'     # Has a bit more orange than 'gold'
        }

        if color in extra_colors:
            color = extra_colors[color]

        self.color_mapping[label] = color




    ##########   The method below require a database connection   ##########


    def prepare_graph(self, result_dataset :[dict], add_edges=True) -> [int]:
        """
        Given a list of dictionary data about graph-database nodes - for example,
        as returned by GraphAccess.get_nodes() - construct and save visualization data for them.

        Each dictionary entry is expected to have a key named "internal_id";
        if not present, it will be silently ignored.
        Though not required, a key named "neo4j_labels" is typically present as well.

        :param result_dataset:  A list of dictionary data about graph-database nodes
        :param add_edges:       If True, all existing edges among the displayed nodes
                                    will also be part of the visualization
        :return:                A list of integers with the internal databased IDs
                                    of all the nodes added to the graph structure
        """
        node_list = []      # Running list of internal databased IDs, for nodes included in visualization
        for node in result_dataset:
            internal_id = node.get("internal_id")
            if not internal_id:
                continue    # Silently ignore any entry lacking this data

            node_list.append(internal_id)

            if "neo4j_labels" in node:
                labels = node["neo4j_labels"]
                del node["neo4j_labels"]
            else:
                labels = ""

            self.add_node(node_id=internal_id, labels=labels,
                          data=node)


        if add_edges:
            # Search the database for any edges among any of the nodes selected for the visualization
            q = '''
                MATCH (n1)-[r]->(n2) 
                WHERE ID(n1) IN $node_list AND ID(n2) IN $node_list 
                RETURN DISTINCT id(n1) AS from_node, id(n2) AS to_node, type(r) AS name
                '''

            result = self.db.query(q, {"node_list": node_list})
            for edge in result:
                #print(edge)
                self.add_edge(from_node=edge["from_node"], to_node=edge["to_node"], name=edge["name"])

        return node_list



    def link_node_groups(self, group1 :[int], group2 :[int]) -> None:
        """
        Search the database for any edges from any of the nodes in the 1st group, to any node in the 2nd group.
        Any located edge will be added to the visualization data stored in this object

        :param group1:  List of integers with the internal databased IDs of the 1st group of nodes
        :param group2:  List of integers with the internal databased IDs of the 2nd group of nodes
        :return:        None
        """
        q = '''
            MATCH (n1)-[r]->(n2) 
            WHERE ID(n1) IN $group1 AND ID(n2) IN $group2 
            RETURN DISTINCT id(n1) AS from_node, id(n2) AS to_node, type(r) AS name
            '''

        result = self.db.query(q, {"group1": group1, "group2": group2})
        for edge in result:
            #print(edge)
            self.add_edge(from_node=edge["from_node"], to_node=edge["to_node"], name=edge["name"])



from life123.visualization.py_graph_visual import PyGraphVisual, PyGraphVisual_OLD
from life123.visualization.graphic_log import GraphicLog, DisplayNetwork
from life123.html_log import HtmlLog as log
from life123.reactions import ReactionUnimolecular, ReactionSynthesis, ReactionDecomposition, ReactionEnzyme, ReactionGeneric
from life123.chem_data import ChemData



class ReactionRegistry:
    """
    Manage a list of reactions, and the reaction-specific objects (defined in reactions.py file),
    such as ReactionUnimolecular, ReactionSynthesis, ReactionDecomposition, ReactionGeneric, ReactionEnzyme, etc.

    Instances of this class are typically used by UniformCompartment objects.
    A ReactionRegistry object may be shared by multiple UniformCompartment objects IF the latter
    all make use of ALL the registered reactions  (i.e. no "pick and choose" some of the reactions.)
    """

    def __init__(self, chem_data=None):

        """
        :param chem_data:   [OPTIONAL] Object of type "ChemData"; if not passed, it will get instantiated automatically -
                                and then it can be obtained by means of the method get_chem_data()
        """

        # TODO: consider adding to argument   "=labels=None"
        """
        assert (chem_data is not None) or (labels is not None), \
            "ReactionRegistry() instantiation: exactly one of the arguments `chem_data` or `labels` must be provided"

        assert (chem_data is None) or (labels is None), \
            "ReactionRegistry() instantiation: cannot specify both the arguments `chem_data` and `labels`"

        if labels is not None:
            chem_data = ChemData(labels=labels)
        """

        #assert chem_data is not None, \
            #"ReactionRegistry() instantiation: the arguments `chem_data` must be provided, and cannot be None"

        if chem_data is None:
            self.chem_data = ChemData()
        else:
            self.chem_data = chem_data

        self.reaction_list = []     # List of objects of the various individual reaction classes,
                                    # such as "ReactionGeneric" and "ReactionEnzyme"


        self.active_chemicals = set()   # Set of the labels of the chemicals - not counting pure catalysts - involved
                                        # in any of the registered reactions
                                        # CAUTION: the concept of "active chemical" might change in future versions, where only SOME of
                                        #          the reactions are simulated.  TODO: it might better belong to UniformCompartment

        #self.active_enzymes = set()     # Set of the labels of the enzymes (catalysts) involved
                                        # in any of the registered reactions
                                        # CAUTION: the concept of "active enzyme" might change in future versions, where only SOME of
                                        #          the reactions are simulated.  TODO: it might better belong to UniformCompartment




    def number_of_reactions(self, include_inactive=False) -> int:
        """
        Return the number of registered chemical reactions

        :param include_inactive:    If True, disabled reactions are also included
        :return:                    The number of registered chemical reactions
        """
        if include_inactive:
            return len(self.reaction_list)

        count = 0
        for rxn in self.reaction_list:
            if rxn.active:
                count += 1

        return count



    def get_all_reactions(self):
        """

        :return:    A list of various types of reaction objects,
                        such as ReactionUnimolecular, ReactionSynthesis, etc.
        """
        return self.reaction_list



    def active_reaction_indices(self) -> [int]:
        """
        Return a list of the reaction index numbers of all the active reactions

        :return:    A list of integers, to identify the active reactions by their indices
        """
        l = []
        for i, rxn in enumerate(self.reaction_list):
            if rxn.active:
                l.append(i)

        return l



    def assert_valid_rxn_index(self, index :int) -> None:
        """
        Raise an Exception if the specified reaction index isn't valid

        :param index:   An integer that indexes the reaction of interest (numbering starts at 0)
        :return:        None
        """
        assert self.number_of_reactions() > 0, \
            f"ChemData.assert_valid_rxn_index(): there are no reactions defined yet.  Use add_reaction() to add them first"

        assert (type(index) == int), \
            f"ChemData.assert_valid_rxn_index(): the requested reaction index must be an integer; " \
            f"the provided value ({index}) is of type {type(index)}"

        assert 0 <= index < self.number_of_reactions(), \
            f"ChemData.assert_valid_rxn_index(): the requested reaction index is not the expected range [0 to {self.number_of_reactions() - 1}], inclusive; " \
            f"the value passed was: {index} (there is no reaction whose index is {index})"



    def get_chem_data(self):
        """
        :return:    Object of type "ChemData"
        """
        return self.chem_data



    def get_reaction(self, i :int):
        """
        Return the data structure of the i-th reaction,
        in the order in which reactions were added (numbering starts at 0)

        :param i:   An integer that indexes the reaction of interest (numbering starts at 0)
        :return:    An object of one of the various individual reaction classes,
                    such as "ReactionGeneric" and "ReactionEnz"
        """
        self.assert_valid_rxn_index(i)

        return self.reaction_list[i]



    def get_reactants(self, i :int) -> [(int, str)]:
        """
        Return a list of pairs with details of the reactants of the i-th reaction.
        Each pair represents a "complex" of the reaction reactants

        :param i:   The index (0-based) to identify the reaction of interest
        :return:    A list of pairs of the form (stoichiometry, chem labels)
        """
        rxn = self.get_reaction(i)
        return rxn.extract_reactants()


    def get_reactant_species(self, i :int) -> [(int, str)]:
        """
        Return a list of the reactant species of the i-th reaction.

        :param i:   The index (0-based) to identify the reaction of interest
        :return:    A list of chem labels
        """
        rxn = self.get_reaction(i)
        reactant_complexes = rxn.extract_reactants()
        return [c[1] for c in reactant_complexes]


    def get_reactants_formula(self, i :int) -> str:
        """
        Return a string with a user-friendly form of the left (reactants) side of the reaction formula

        :param i:   The index (0-based) to identify the reaction of interest
        :return:    A string with a user-friendly form of the left (reactants) side of the chemical reaction
        """
        rxn = self.get_reaction(i)
        return rxn.extract_reactants_formula()



    def get_products(self, i :int) -> [(int, str)]:
        """
        Return a list of pairs with details of the products of the i-th reaction.
        Each pair represents a "complex" of the reaction products

        :param i:   The index (0-based) to identify the reaction of interest
        :return:    A list of pairs of the form (stoichiometry, species label)
        """
        rxn = self.get_reaction(i)
        return rxn.extract_products()


    def get_product_species(self, i :int) -> [(int, str)]:
        """
        Return a list of the product species of the i-th reaction.

        :param i:   The index (0-based) to identify the reaction of interest
        :return:    A list of chem labels
        """
        rxn = self.get_reaction(i)
        product_complexes = rxn.extract_products()
        return [c[1] for c in product_complexes]


    def get_products_formula(self, i :int) -> str:
        """
        Return a string with a user-friendly form of the right (products) side of the reaction formula

        :param i:   The index (0-based) to identify the reaction of interest
        :return:    A string with a user-friendly form of the right (products) side of the chemical reaction
        """
        rxn = self.get_reaction(i)
        return rxn.extract_products_formula()



    def get_forward_rate(self, i :int) -> float:
        """
        Return the value of the forward rate constant of the i-th reaction

        :param i:   The integer index (0-based) to identify the reaction of interest
        :return:    The value of the forward rate constant for the above reaction
        """
        rxn = self.get_reaction(i)
        return rxn.extract_forward_rate()


    def get_reverse_rate(self, i :int) -> float:
        """
        Return the value of the reverse (back) rate constant of the i-th reaction

        :param i:   The integer index (0-based) to identify the reaction of interest
        :return:    The value of the reverse (back) rate constant for the above reaction
        """
        rxn = self.get_reaction(i)
        return rxn.extract_reverse_rate()



    def get_chemicals_in_reaction(self, rxn_index :int) -> {int}:
        """
        Return a SET of indices (being a set, they're NOT in any particular order)
        of all the chemicals participating in the i-th reaction

        :param rxn_index:   An integer with the (zero-based) index to identify the reaction of interest
        :return:            A SET of indices of the chemicals involved in the above reaction
                                Note: being a set, it's NOT in any particular order
        """
        rxn = self.get_reaction(rxn_index)

        name_set = rxn.extract_chemicals_in_reaction()

        index_set = {self.chem_data.get_index(name) for name in name_set}

        return index_set


    def get_chemicals_indexes_in_reaction(self, rxn_index :int) -> [int]:
        """
        Return a sorted list of the indexes
        of all the chemicals participating in the i-th reaction

        :param rxn_index:   An integer with the (zero-based) index to identify the reaction of interest
        :return:            A sorted list of indices of the chemicals involved in the above reaction
        """
        index_set = self.get_chemicals_in_reaction(rxn_index)
        index_list = list(index_set)

        return sorted(index_list)



    def get_reactions_participating_in(self, chem_label :str, side :str) -> list:
        """
        Return a list of all the reactions that the given chemical species
        is involved in

        :param chem_label:
        :param side:        Either "left" or "right"
        :return:            List of varius types of "Reaction" objects
        """
        # TODO: test
        assert side == "left" or side == "right"
        rxns_found_in = []
        for rxn in self.reaction_list:
            if side == "left":
                if chem_label in rxn.extract_reactant_labels():
                    rxns_found_in.append(rxn)
            else:
                 if chem_label in rxn.extract_product_labels():
                    rxns_found_in.append(rxn)


        return rxns_found_in




    def _parse_reaction_term(self, term :str|tuple|list, name="term") -> (int, str):
        """
        Accept various ways to specify a reaction term, and return a standardized triplet form for it.

        NOTE: if the stoichiometry coefficient isn't specified, it defaults to 1

        In the passed tuples or lists:
            - required 1st entry is the stoichiometry
            - required 2nd entry is the chemical name

        If just a string is being passed, it is taken to be the chemical name,
        with stoichiometry coefficient of 1

        EXAMPLES:
            "F"          gets turned into:  (1, "F")   - defaults used for stoichiometry
            (2, "F")                        (2, "F")
            It's equally acceptable to use LISTS in lieu of tuples

        :param term:    A string (a chemical name)
                            OR  a pair (stoichiometry coeff, name)
        :param name:    An optional nickname, handy to refer to this term in error messages if needed
                            (for example, "reactant" or "product")
        :return:        A standardized pair of the form (stoichiometry, species_label),
                            where stoichiometry is an integer, while species_label is a string
        """
        if type(term) == str:
            return  (1, term)    # Accept simply the chemical name as a shortcut,
                                    # for when the stoichiometry coefficient and reaction order are both 1

        if type(term) != tuple and type(term) != list:
            raise Exception(f"_parse_reaction_term(): {name} must be either a string (a chemical name), "
                            f"or a pair (stoichiometry coeff, name). "
                            f"Instead, it is `{term}` (of type {type(term)})")

        # If we get thus far, term is either a tuple or a list
        assert len(term) == 2,  \
            f"_parse_reaction_term(): Unexpected length for {name} tuple/list: it should be 2. " \
            f"Instead, it is {len(term)}"

        stoichiometry = term[0]
        assert type(stoichiometry) == int, \
            f"_parse_reaction_term(): The stoichiometry coefficient must be an integer. Instead, it is {stoichiometry}"

        chem_label = term[1]
        assert type(chem_label) == str, \
                            f"_parse_reaction_term(): The chemical name must be a string. " \
                            f"Instead, it is `{chem_label}` (of type {type(chem_label)})"


        return (stoichiometry, chem_label)



    def _standardize_reaction_side(self, terms :str|list, arg_name :str):
        """

        :param terms:
        :param arg_name:
        :return:
        """
        assert terms is not None, \
            f"standardize_reaction_side(): the argument `{arg_name}` is a required one; it can't be None"

        if type(terms) == str:
            terms = [terms]
        else:
            assert type(terms) == list, \
                f"standardize_reaction_side(): the argument `{arg_name}` must be a string or a list; the passed value was {type(terms)}"

        return [self._parse_reaction_term(r, "reactant") for r in terms]   # A list of pairs



    def add_elementary_reaction(self, reactants :str|list, products :str|list,
                                reaction_type=None, temp=None,
                                **kwargs) -> int:
        """
        Create and register a new SINGLE elementary chemical reaction,
        optionally including its kinetic and/or thermodynamic data.
        All the involved chemicals can be either previously registered, or not.

        :param reactants:       A string or pair of strings; for reactions such as 2 A -> P, pass ["A", "A"]
        :param products:        A string or pair of strings; for reactions such as A -> 2 P, pass ["P", "P"]
        :param reaction_type:   A string with one of the following values:
                                    "ReactionUnimolecular", "ReactionSynthesis", "ReactionDecomposition"

        :param temp:            [OPTIONAL] Temperature in Kelvins
        :param kwargs:          [OPTIONAL] Other named arguments to pass to instantiate the various reaction
                                objects, such as `ReactionSynthesis`.
                                For list, see documentation of the classes in reactions.py

        :return:                Integer index of the newly-added reaction
                                    (in the list self.reaction_list, stored as object variable)
        """
        # Turn `reactants` and `products` into lists, if not already lists
        if type(reactants) == str:
            reactants = [reactants]
            n_reactants = 1
        else:
            n_reactants = len(reactants)
            assert n_reactants <= 2

        if type(products) == str:
            products = [products]
            n_products = 1
        else:
            n_products = len(products)
            assert n_products <= 2


        if reaction_type is None:
            # Determine the reaction type from the number of reactants and products
            if n_reactants == 1 and n_products == 1:
                reaction_type = "ReactionUnimolecular"
            elif n_reactants == 2 and n_products == 1:
                reaction_type = "ReactionSynthesis"
            elif n_reactants == 1 and n_products == 2:
                reaction_type = "ReactionDecomposition"
            else:
                raise Exception(f"add_elementary_reaction(): {n_reactants} reactants and {n_products} products cannot correspond to any elementary reaction")


        match reaction_type:
            case "ReactionUnimolecular":
                rxn = ReactionUnimolecular(reactant=reactants[0], product=products[0], temp=temp, **kwargs)
            case "ReactionSynthesis":
                rxn = ReactionSynthesis(reactants=reactants, product=products[0], temp=temp, **kwargs)
            case "ReactionDecomposition":
                rxn = ReactionDecomposition(reactant=reactants[0], products=products, temp=temp, **kwargs)
            case _:
                raise Exception(f"add_elementary_reaction(): unknown reaction type ({reaction_type})")


        #print(f"add_elementary_reaction(): adding reaction of type `{reaction_type}`")

        return self.register_reaction(rxn=rxn, temp=temp)



    def add_reaction(self, reactants :str|list, products :str|list,
                     kF=None, kR=None,
                     enzyme=None, k1_F=None, k1_R=None, k2_F=None,
                     temp=None,
                     **kwargs) -> int:
        """
        Create and register a new SINGLE chemical reaction,
        optionally including its kinetic and/or thermodynamic data.
        All the involved chemicals can be either previously registered, or not.

        NOTE: in the reactants and products, if the stoichiometry coefficients aren't specified,
              they're assumed to be 1.

              The full structure of each term in the list of reactants and of products
              is the pair:  (stoichiometry coefficient, chemical label)

              EXAMPLES of formats to use for each term in the lists of the reactants and of the products:
                "F"         is taken to mean (1, "F") - default stoichiometry
                (2, "F")    is taken to mean (2, "F") - stoichiometry coefficient specified

              It's equally acceptable to use LISTS in lieu of tuples for the pairs

        :param reactants:       A string or list of pairs (stoichiometry, species name),
                                    or simplified terms in various formats; for details, see above
        :param products:        A string or list of pairs (stoichiometry, species name),
                                    or simplified terms in various formats; for details, see above

        :param temp:            [OPTIONAL] Temperature in Kelvins
        :param kwargs:          [OPTIONAL] Other named arguments to pass to instantiate the various reaction
                                objects, such as `ReactionSynthesis`.
                                For list, see documentation of the classes in reactions.py

        :return:                Integer index of the newly-added reaction
                                    (in the list self.reaction_list, stored as object variable)
        """
        #TODO: add an optional `reaction_type` argument;
        #      the dispatching might best be done elsewhere

        # Determine and assign the specific type of reaction, and instantiate an object of that class
        if enzyme is not None:
            reaction_type = "ReactionEnzyme"
            assert type(reactants) == str
            assert type(products) == str
            rxn = ReactionEnzyme(enzyme=enzyme, substrate=reactants, product=products,
                                 k1_F=k1_F, k1_R=k1_R, k2_F=k2_F, temp=temp, **kwargs)
        else:
            reactant_list = self._standardize_reaction_side(reactants, arg_name="reactants")
            product_list = self._standardize_reaction_side(products, arg_name="products")

            single_reactant = None
            if len(reactant_list) == 1 and reactant_list[0][0] == 1:    # A single reactant, with stoichiometry 1
                single_reactant = reactant_list[0][1]

            single_product = None
            if len(product_list) == 1 and product_list[0][0] == 1:      # A single product, with stoichiometry 1
                single_product = product_list[0][1]

            reaction_type = "ReactionGeneric"       # Default value, possibly changed below

            if single_reactant:    # A single reactant, with stoichiometry 1
                if single_product:      # A single product, with stoichiometry 1
                    reaction_type = "ReactionUnimolecular"
                    rxn = ReactionUnimolecular(reactant=single_reactant, product=single_product, kF=kF, kR=kR, temp=temp, **kwargs)
                elif len(product_list) == 2 and product_list[0][0] == 1 and product_list[1][0] == 1:      # Two products, both with stoichiometry 1
                    reaction_type = "ReactionDecomposition"
                    rxn = ReactionDecomposition(reactant=single_reactant, products=[product_list[0][1], product_list[1][1]],
                                                kF=kF, kR=kR, temp=temp, **kwargs)
                elif len(product_list) == 1 and product_list[0][0] == 2:      # A product with stoichiometry 2  (EXAMPLE : A <-> 2 B)
                    reaction_type = "ReactionDecomposition"
                    rxn = ReactionDecomposition(reactant=single_reactant, products=[product_list[0][1], product_list[0][1]],
                                                kF=kF, kR=kR, temp=temp, **kwargs)
            elif single_product:
                if len(reactant_list) == 2 and reactant_list[0][0] == 1 and reactant_list[1][0] == 1:      # Two reactants, both with stoichiometry 1
                    reaction_type = "ReactionSynthesis"
                    rxn = ReactionSynthesis(reactants=[reactant_list[0][1], reactant_list[1][1]],
                                            product=single_product, kF=kF, kR=kR, temp=temp, **kwargs)
                elif len(reactant_list) == 1 and reactant_list[0][0] == 2:  # A reactant with stoichiometry 2  (EXAMPLE : 2A <-> P)
                    reaction_type = "ReactionSynthesis"
                    rxn = ReactionSynthesis(reactants=[reactant_list[0][1], reactant_list[0][1]],
                                            product=single_product, kF=kF, kR=kR, temp=temp, **kwargs)

            if reaction_type == "ReactionGeneric":
                 rxn = ReactionGeneric(reactants=reactant_list, products=product_list, kF=kF, kR=kR, temp=temp, **kwargs)


        print(f"add_reaction(): detected reaction type `{reaction_type}`")

        return self.register_reaction(rxn=rxn, temp=temp)



    def register_reaction(self, rxn, temp=None) -> int:
        """
        Register a SINGLE chemical reaction from its reaction-specific object,
        and set all its kinetic and/or thermodynamic data from the available information,
        including the value of the temperature (stored in object variable.)

        All the involved chemicals can be either previously registered, or not;
        if not, they will get automatically registered.

        :param rxn: Object of one of the specific Reaction classes, such as
                        ReactionUnimolecular, ReactionSynthesis, ReactionDecomposition,
                        ReactionEnz, ReactionGeneric
        :param temp:Temperature in degree Kelvin

        :return:    Integer index of the newly-added reaction
                        (in the list self.reaction_list, stored as object variable)
        """
        self.reaction_list.append(rxn)

        # Register any newly-encountered reactant not already registered
        # for aesthetic reasons, we'll do 1) catalyst, 2) reactants, 3) products

        #catalyst = rxn.extract_catalyst()

        #if catalyst:
            # Register the catalyst, if not already registered
            #self.chem_data.add_chemical(name=catalyst, skip_duplicates=True)

        # Register any newly-encountered reactant not already registered
        rxn_reactants = rxn.extract_reactant_labels()
        for label in rxn_reactants:
            self.chem_data.add_chemical(name=label, skip_duplicates=True)

        # Register any newly-encountered reaction product not already registered
        rxn_products = rxn.extract_product_labels()
        for label in rxn_products:
            self.chem_data.add_chemical(name=label, skip_duplicates=True)

        # Register any newly-encountered reaction intermediates not already registered
        rxn_intermediate = rxn.extract_intermediate()
        if rxn_intermediate:
            new_index = self.chem_data.add_chemical(name=rxn_intermediate, skip_duplicates=True)
            if new_index is not None:
                print(f"register_reaction() INFO: a reaction intermediates (`{rxn_intermediate}`), not explicitly registered, was automatically added to the chemical registry")

            if self.chem_data.get_diffusion_rate(chem_label=rxn_intermediate) is None:
                # Attempt to estimate the diffusion rate constant of the reaction intermediate
                D_enzyme = self.chem_data.get_diffusion_rate(chem_label=rxn_intermediate)
                D_substrate = self.chem_data.get_diffusion_rate(chem_label=rxn.substrate)
                # TODO: this might best belong elsewhere
                if (D_enzyme is not None) and (D_substrate is not None):
                    D_ES_rough_estimate = min(D_enzyme, D_substrate) * 0.9
                    print(f"register_reaction() INFO: diffusion rate for the reaction intermediates (`{rxn_intermediate}`), not yet specified, roughly estimated as {D_ES_rough_estimate}")
                    self.chem_data.set_diffusion_rate(chem_label=rxn_intermediate, diff_rate=D_ES_rough_estimate)


        involved_chemicals = set(rxn_reactants) | set(rxn_products)  # Union of sets
        if  rxn_intermediate:
             involved_chemicals = involved_chemicals | {rxn_intermediate}     # Union of sets

        #if catalyst is not None:
            #involved_chemicals = involved_chemicals - {catalyst}    # Difference between sets
            #self.active_enzymes.add(rxn.catalyst)                   # Add the new entry to a set

        # Update the set of "active chemicals"
        self.active_chemicals = self.active_chemicals | involved_chemicals  # Union of sets

        if temp is not None:
            rxn.set_thermodynamic_data(temp=temp)   # TODO: unclear if this is the best place to do this

        return len(self.reaction_list) - 1



    def clear_reactions_data(self) -> None:
        """
        Get rid of all the reactions; start again with "an empty slate" (but still with reference
        to the same data object about the chemicals and their properties)

        :return:    None
        """
        self.reaction_list = []
        self.active_chemicals = set()
        #self.active_enzymes = set()



    def inactivate_reaction(self, i :int) -> None:
        """
        Mark the i-th reaction as "inactive/disabled" (essentially, "deleted", but holding its positional
        index, to avoid a change in index in other reactions)

        TODO: Not yet supported by the dynamical modules; DON'T USE YET in simulations!

        :param i:   Zero-based index of the reaction to disable
        :return:    None
        """
        rxn = self.get_reaction(i)
        rxn.active = False

        # Re-construct self.active_chemicals and self.active_enzymes
        self.active_chemicals = set()
        #self.active_enzymes = set()

        for rxn in self.reaction_list:
            involved_chemicals = rxn.extract_chemicals_in_reaction()
            involved_chemicals = involved_chemicals - {rxn.catalyst}        # Set difference
            self.active_chemicals = self.active_chemicals.union(involved_chemicals)     # Union of sets
            #if rxn.catalyst is not None:
                #self.active_enzymes.add(rxn.catalyst)       # Add the new entry to a set





    #####################################################################################################

    '''                             ~   TO DESCRIBE THE REACTIONS  ~                                  '''

    def ________DESCRIBE_REACTIONS________(DIVIDER):
        pass        # Used to get a better structure view in IDEs
    #####################################################################################################


    def describe_reactions(self, concise=False) -> None:
        """
        Print out a user-friendly plain-text form of ALL the reactions.
        If wanting to describe just 1 reaction, use single_reaction_describe()

        EXAMPLE (not concise):
            Number of reactions: 2
            (0) CH4 + 2 O2 <-> CO2 + 2 H2O  (kF = 3.0 / kR = 2.0 / Delta_G = -1,005.13 / K = 1.5) | 1st order in all reactants & products
            (1) A + B <-> C  (kF = 5.0 / kR = 1.0 / Delta_G =  / K = 5.0) | 1st order in all reactants & products
            Chemicals involved in the above reactions: {'CH4', 'O2', 'H2O', 'A', 'B', 'C'}

        :param concise:     If True, less detail is shown
        :return:            None
        """
        #TODO: implement a Pandas version
        print(f"Number of reactions: {self.number_of_reactions()}")

        # Print a concise description of EACH REACTION IN TURN
        for description in self.multiple_reactions_describe(concise=concise):
            print(description)

        chem_labels = self.labels_of_active_chemicals(sort_by_index=True)   # Set of chem labels, sorted by chemical index

        if self.chem_data.color_dict != {}:   # If plot colors were registered, show them alongside the chem labels
            chem_labels_with_colors = []
            for label in chem_labels:
                color = self.chem_data.get_plot_color(label)
                if color:
                    chem_labels_with_colors.append(f'"{label}" ({color})')
                else:
                    chem_labels_with_colors.append(f'"{label}"')

            chem_labels = "{" + ", ".join(chem_labels_with_colors) + "}"


        #if self.active_enzymes == set():    # If no enzymes were involved in any reaction
        print(f"Chemicals involved in the above reactions: {chem_labels}")
        '''
        else:
            print(f"Chemicals involved in the above reactions (not counting enzymes): {chem_labels}")

            print(f"Enzymes involved in the above reactions: ")
            for enz in self.names_of_enzymes():
                enzyme_color = self.chem_data.get_plot_color(enz)
                if enzyme_color:
                    print(f'  "{enz}" ({enzyme_color})')
                else:
                    print(f'  "{enz}"')
        '''



    def multiple_reactions_describe(self, rxn_list=None, concise=False) -> [str]:
        """
        The counterpart of single_reaction_describe() for many reactions.
        Return a list of strings, each string being a (concise or not) user-friendly plain-text form of
        each of the reactions

        :param rxn_list:    Either a list of integers, to identify the reactions of interest,
                                or None, meaning ALL reactions
        :param concise:     If True, less detail is shown
        :return:            A list of strings; each string is the description of one of the reactions
        """
        if rxn_list is None:
            rxn_list = range(self.number_of_reactions())    # Show ALL reactions, by default

        out = []    # Output list being built

        for i in rxn_list:
            description = self.single_reaction_describe(rxn_index=i, concise=concise)
            description = f"{i}: {description}"
            out.append(description)

        return out



    def single_reaction_describe(self, rxn_index: int, concise=False) -> str:
        """
        Return as a string, a user-friendly plain-text form of the given reaction
        EXAMPLE (concise):      "CH4 + 2 O2 <-> CO2 + 2 H2O"
        EXAMPLE (not concise):  "CH4 + 2 O2 <-> CO2 + 2 H2O  (kF = 3.0 / kR = 2.0 / Delta_G = -1,005.13 / K = 1.5) | 1st order in all reactants & products"

        :param rxn_index:   Integer to identify the reaction of interest
        :param concise:     If True, less detail is shown
        :return:            A string with a description of the specified reaction
        """
        rxn = self.get_reaction(rxn_index)

        return rxn.describe(concise)    # Invoke the individual reaction object



    def labels_of_active_chemicals(self, sort_by_index=False) -> [str]:
        """
        Return a list of the labels of all the chemicals
        involved in ANY of the registered reactions,
        but NOT counting chemicals that always appear
        in a catalytic role in all the reactions they participate in
        (if a chemical participates in a non-catalytic role in ANY reaction, it'll appear here)

        The list is not in any particular order, unless sort_by_index is True

        :param sort_by_index:   If True, the list is sorted by the index (order of registration)
                                    of the chemicals in it
        :return:                A set of chemical labels
        """
        if not sort_by_index:
            return list(self.active_chemicals)

        return sorted(self.active_chemicals, key=self.chem_data.get_index)



    def number_of_active_chemicals(self) -> int:
        """
        Return the number of all the chemicals
        involved in ANY of the registered reactions,
        but NOT counting chemicals that always appear
        in a catalytic role in all the reactions they participate in
        (if a chemical participates in a non-catalytic role in ANY reaction, it'll appear here)
        """
        return len(self.active_chemicals)



    def indexes_of_active_chemicals(self) -> [int]:
        """
        Return the ordered list (numerically sorted) of the INDEX numbers of all the chemicals
        involved in ANY of the registered reactions,
        but NOT counting chemicals that always appear in a catalytic role in all the reactions they
        participate in
        (if a chemical participates in a non-catalytic role in ANY reaction, it'll appear here.)

        EXAMPLE: [2, 7, 8]  if only those 3 chemicals (with indexes of, respectively, 2, 7 and 8)
                            are actively involved in ANY of the registered reactions

        CAUTION: the concept of "active chemical" might change in future versions, where only SOME of
                 the reactions are simulated
        """
        index_list = list(map(self.chem_data.get_index, self.active_chemicals))
        return sorted(index_list)





    #####################################################################################################

    '''                          ~   FOR CREATION OF NETWORK DIAGRAMS  ~                              '''

    def ________NETWORK_DIAGRAMS________(DIVIDER):
        pass        # Used to get a better structure view in IDEs
    #####################################################################################################


    def prepare_graph_network(self) -> dict:
        """
        Prepare and return a data structure with chemical-reaction data in a network format,
        ready to be passed to the front end, for network-diagram visualization with the Cytoscape.js library
        (in the graph module "vue_cytoscape")

        4 parts are generated, and assembled together as a dictionary with 4 keys: 'nodes', 'edges', 'color_mapping', 'caption_mapping'

        EXAMPLE of the 'nodes' structure part of the returned object for an  A <-> B reaction:
           [{'id': 'C-0', 'labels': ['Chemical'], 'name': 'A', 'diff_rate': None},
            {'id': 'C-1', 'labels': ['Chemical'], 'name': 'B', 'diff_rate': None},

            {'id': 'R-0', 'labels': ['Reaction'], 'name': 'RXN', 'kF': 3.0, 'kR': 2.0, 'K': 1.5, 'Delta_G': -1005.13}
           ]

        EXAMPLE of the 'edges' structure part of the returned object for that same  A <-> B reaction:
           [
            {'id': 'edge-1', 'name': 'produces', 'source': 'R-0', 'target': 'C-1', 'stoich': 1, 'rxn_order': 1},
            {'id': 'edge-2', 'name': 'reacts',   'source': 'C-0', 'target': 'R-0', 'stoich': 1, 'rxn_order': 1}
           ]

        EXAMPLE of `color_mapping`:     {'Chemical': '#8DCC92', 'Reaction': '#D9C8AD'},
        EXAMPLE of `caption_mapping`:   {'Chemical': 'name', 'Reaction': 'name'}}

        :return:    A dictionary with 4 keys: 'nodes', 'edges', 'color_mapping', 'caption_mapping'
        """
        graph = PyGraphVisual()     # Object to facilitate data preparation for graph visualization

        # Note: the graph nodes representing Chemicals will be given an id such as "C-123" and a label "Chemical";
        #       the graph nodes representing Reactions will be given an id such as "R-456" and a label "Reaction"

        for i, rxn in enumerate(self.reaction_list):    # Consider each REACTION in turn
            # Add a node representing the reaction
            rxn_id = f"RXN-{i}"               # Example: "RXN-456"
            node_data = {'name': 'RXN', 'formula': rxn.describe(concise=True)}

            # Show the parameter of the reaction
            rxn_properties = rxn.extract_rxn_properties()
            for k,v in rxn_properties.items():
                node_data[k] = f"{v:,.6g}"

            graph.add_node(node_id=rxn_id, labels='Reaction', properties=node_data)


            # Process all the PRODUCTS of this reaction
            products = rxn.extract_products()
            for term in products:
                species_name = rxn.extract_chem_label(term)
                chemical_id = f"C-{self.chem_data.get_index(species_name)}"      # Example: "C-12"

                # Add each product to the graph as a node (if not already present)
                properties={'name': species_name}
                if diff := self.chem_data.get_diffusion_rate(chem_label=species_name):
                    properties['diff_rate'] = diff
                graph.add_node( node_id=chemical_id, labels="Chemical",
                                properties=properties)

                # Append edge from "reaction node" to "product node"
                graph.add_edge(from_node=rxn_id, to_node=chemical_id, name="produces",
                               properties={'stoich': rxn.extract_stoichiometry(term)})


            # Process all the REACTANTS of this reaction
            reactants = rxn.extract_reactants()
            for term in reactants:
                species_name = rxn.extract_chem_label(term)
                chemical_id = f"C-{self.chem_data.get_index(species_name)}"      # Example: "C-34"

                # Add each reactant to the graph as a node (if not already present)
                properties={'name': species_name}
                if diff := self.chem_data.get_diffusion_rate(chem_label=species_name):
                    properties['diff_rate'] = diff
                graph.add_node(node_id=chemical_id, labels="Chemical",
                               properties=properties)

                # Append edge from "reactant node" to "reaction node"
                graph.add_edge(from_node=chemical_id, to_node=rxn_id, name="reacts",
                               properties={'stoich': rxn.extract_stoichiometry(term)})


        graph.assign_color_mapping(label='Chemical', color='graph_green')
        graph.assign_color_mapping(label='Reaction', color='graph_lightbrown')

        graph.assign_caption(label='Chemical', caption='name')
        graph.assign_caption(label='Reaction', caption='id')

        #print(graph)

        return graph.get_graph_data()



    def plot_reaction_network(self, log_file :str, graphic_component="vue_cytoscape_5") -> None:
        """
        Send a plot of the network of reactions to the HTML log file,
        also including a brief summary of all the reactions.

        :param log_file:            The name of the file into which to place the HTML code
                                        to create the interactive network plot.
                                        The suffix ".htm" will be added if it doesn't end with ".htm" or ".html"
                                        If the file already exists, it will get overwritten.
                                        (Note: this file will automatically include an internal reference to the JavaScript
                                        file specified in `graphic_component`)
        :param graphic_component:   The name of a Vue component that accepts a "graph_data" argument,
                                        an object with the following keys
                                        'nodes', 'edges', 'color_mapping' and 'caption_mapping
                                        For more details, see prepare_graph_network()

        :return:                    None
        """
        assert graphic_component == "vue_cytoscape_5", \
            "plot_reaction_network(): the only value supported for argument `graphic_component` is 'vue_cytoscape_5'"

        # Send a brief summary of all the reactions to the HTML log file
        header = "<h3>List of reactions:</h3>\n"

        header += "<pre>"    #"<p style='font-family: monospace; padding-left:20px'>"
        rxn_descriptions = self.multiple_reactions_describe(concise=True)
        for desc in rxn_descriptions:
            header += f"    {desc}\n"
        header += "</pre>\n"

        graph_data = self.prepare_graph_network()
        # A dictionary with 4 keys: ''nodes', 'edges', 'color_mapping' and 'caption_mapping'

        # Send a plot of the network of reactions to the HTML log file
        #GraphicLog.export_plot(graph_data, graphic_component, unpack=unpack)
        DisplayNetwork.export_plot(graph_data = graph_data,
                                   graphic_component = graphic_component,
                                   filename = log_file, caption=header)



from life123.chem_data import ChemData
from life123.reaction_registry import ReactionRegistry
import numpy as np



class RandomReactionNetwork:
    """
    EARLY PARTIAL IMPLEMENTATION!

    Generate and manage random networks of reactions that are thermodynamically possible,
    and biologically plausible.

    Kinetic parameters NOT yet implemented.
    """

    def __init__(self, n_species :int, n_rxns :int, relative_rxn_prob=None, seed=None, temp=298.15, verbose=False):
        """
        :param n_species:           Integer >=3, with the desired number of chemicals species
        :param n_rxns:              Number of desired reactions (ok to use 0 for testing)
        :param relative_rxn_prob:   [OPTIONAL] A 3-element list with the relative probability
                                        of each of the 3 reaction choices that are currently available:
                                        [ReactionUnimolecular, ReactionSynthesis, ReactionDecomposition]
                                        Probabilities must add up to 1;
                                        if not specified, a uniform distribution is used.
                                        EXAMPLE:  [0.2, 0.5, 0.3]
        :param seed:                [OPTIONAL] A (large) integer, to guarantee consistent random runs
        :param temp:                [OPTIONAL] Temperature in Kelvins
        :param verbose:             [OPTIONAL] Some extra printout if True
        """
        RXN_CHOICES = ["ReactionSynthesis", "ReactionDecomposition"]  # Available reaction classes
        #RXN_CHOICES = ["ReactionUnimolecular", "ReactionSynthesis", "ReactionDecomposition"]

        assert type(n_species) == int, \
            f"RandomReactionNetwork Instantiation: the argument `n_chem` must be an integer, not {type(n_species)}"

        assert n_species >= 3, \
            f"RandomReactionNetwork Instantiation: the number of chemicals must be at least 3 (value passed was {n_species})"

        assert type(n_rxns) == int, \
            f"RandomReactionNetwork Instantiation: the argument `n_rxns` must be an integer, not {type(n_rxns)}"


        self.chem_data = ChemData(n_chems=n_species)        # Object of type "ChemData", with info on the individual chemicals

        all_chems = self.chem_data.get_all_labels()
        # Auto-generated strings "A", "B", ..., "Z", "Z2", "Z3"

        self.registry = ReactionRegistry(chem_data=self.chem_data)      # Object of type ReactionRegistry,
                                                                        # to manage a list of reactions, and the reaction-specific objects

        # A Generator object with Numpy’s default BitGenerator
        self.rng = np.random.default_rng(seed=seed)

        # Select each of the reactions types (using the relative probabilities, if available)
        rxn_type = self.rng.choice(a=RXN_CHOICES, size=n_rxns, replace=True, p=relative_rxn_prob)  # List of class names
        # EXAMPLE: [ReactionUnimolecular, ReactionSynthesis, ReactionDecomposition]

        self.standard_species_enthalpy = {}

        self.temp = temp


        MAX_ATTEMPTS = 6        # To avoid infinite loops
        for r in rxn_type:
            # For each reaction to create
            #print(f"Attempting to add reaction of type: {r}")
            reactants, products = self._assign_chems_to_rxn(r, all_chems)
            #print(f"    Tentatively considering reactants: {reactants} | products: {products}")

            n_attempts = 1
            # Prevent re-adding reactions that identically occurred before
            while n_attempts <= MAX_ATTEMPTS and self.already_used(reactants, products):
                reactants, products = self._assign_chems_to_rxn(r, all_chems)
                #print(f"    the previous combination had already been used.  Tentatively considering new "
                #      f"reactants: {reactants} | products: {products}")
                n_attempts += 1
            assert n_attempts <= MAX_ATTEMPTS, \
                f"Giving up after {n_attempts} attempts.  It seems that you are requesting such a large number of reactions ({n_rxns}), " \
                f"relative to the {n_species} available chemicals, that random reactions would be repeating"


            delta_enthalpy = self.random_reaction_enthalpy(reactants, products)
            delta_entropy = self.random_reaction_entropy(reaction_type=r)
            i = self.registry.add_elementary_reaction(reactants=reactants, products=products,
                                                      delta_H=delta_enthalpy, delta_S=delta_entropy,
                                                      temp=self.temp)

            if verbose:
                print(f"-- Added reaction {i}, with reactants: {reactants} | products: {products} | delta_enthalpy: {delta_enthalpy}")
                print(self.registry.get_reaction(i).describe(concise=False))



    def get_reactions(self):
        """
        Extract and return all the reactions

        :return:    Object of type "ReactionRegistry"
        """
        return self.registry



    def _assign_chems_to_rxn(self, reaction_type :str, all_chem_labels :[str]) -> (list, list):
        """
        From the pool of available chemicals, randomly assign reactants and products to
        the specified type of reaction.

        Note: validation checks are assumed to be done in the calling function

        :param reaction_type:   One of the following reaction types:
                                    "ReactionSynthesis", "ReactionDecomposition"
        :param all_chem_labels: All the chemical labels for the random network we're building
        :return:                The pair (reactants, products),
                                    where each element is a list of chemical labels
                                    EXAMPLE (for a synthesis reaction):    (['C', 'C'], ['A'])
        """
        n_reactants = 1
        n_products = 1

        if reaction_type == "ReactionDecomposition":
            n_products = 2      # The only scenario leading to 2 products
        elif reaction_type == "ReactionSynthesis":
            n_reactants = 2     # The only scenario leading to 2 reactants
        else:
            raise Exception(f"Only `ReactionDecomposition` and `ReactionSynthesis` are currently allowed; "
                            f"you requested `{reaction_type}`")


        # Random selection of reactants are WITH replacement, to allow reactions such as 2 A -> B
        reactants = self.rng.choice(a=all_chem_labels, size=n_reactants, replace=True)

        # Likewise, random selection of products are with replacement, to allow reactions such as A -> 2 B
        # However, do NOT pick the products from the pool of reactants that were used in this reaction!
        available_pool = [c for c in all_chem_labels
                             if c not in reactants]         # EXAMPLE: ['X', 'Y']
        #print(available_pool)

        products = self.rng.choice(a=available_pool, size=n_products, replace=True)

        reactants = reactants.tolist()  # Converts both the container (numpy.ndarray) and the element types (numpy.str_),
                                        # so as to get a python list of strings
        products = products.tolist()

        return reactants, products



    def already_used(self, reactants :[str], products :[str]) -> bool:
        """
        Check whether the given set of reactants, and the given set of products,
        have already been used in a previously-generated random reaction,
        or in its reverse reaction.

        The stoichiometry is not taken into account because, for example, if we already
        utilized D -> C , we do not want to generate D -> 2 C ,
        and thus regard it as "already used".

        ~~~ EXAMPLES ~~~
        If we previous generated  A -> B  then:
            already_used(self, reactants="A", products="B") will be True
            already_used(self, reactants="B", products="A") will be True
            already_used(self, reactants=["A", "A'], products="B") will be True

        :param reactants:   A list of the labels of the reactants in the reaction to look up
        :param products:    A list of the labels of the products in the reaction to look up
        :return:            True if found, or False if not
        """
        reactant_set = set(reactants)
        product_set = set(products)

        for rxn in self.registry.reaction_list:
            if set(rxn.extract_reactant_labels()) == reactant_set \
                    and set(rxn.extract_product_labels()) == product_set:
                return True

            # Check the reverse reaction as well
            if set(rxn.extract_reactant_labels()) == product_set \
                    and set(rxn.extract_product_labels()) == reactant_set:
                return True

        return False



    def random_species_enthalpy(self, sigma = 40.4145, n=None):
        """
        A random standard enthalpy value to assign to each hypothetical chemical.
        Essentially, a essentially a generalized, randomized version of enthalpy of formation
        for n chemical species.

        The values are generated from a normal distribution all centered at zero,
        and with the passed value for the standard deviation sigma.

        The default value for sigma has the effect of making the derived SD
        of the sum of 3 random values to be about
        sqrt(3) * sigma = 70 kJ/mol
        For example, the sum of 3 random values comes up in determining
        the reaction enthalpy changes in reactions such as  A + B <-> C ,
        as computed by random_reaction_enthalpy()
        [But in the case of reactions such as 2 A <-> C, the reaction enthalpy change
        would be sqrt(5) * sigma = 90.4 kJ/mol]

        :param sigma:   [OPTIONAL] The standard deviation of the (zero-centered)
                            normal distribution used to generate all values.
                            Default: 40.4145 kJ/mol
        :param n:       [OPTIONAL] If not provided, return a single number;
                            otherwise, return a list of n values.
                            All values are kJ/mol
        :return:
        """
        return self.rng.normal(loc=0, scale=sigma, size=n)



    def get_species_enthalpy(self, label :str):
        """
        Look up the standard enthalpy value to assign to the given chemical;
        if not found, assign and store a new random one

        :param label:   To identify a chemical species
        :return:        A standard enthalpy value already assigned, or just assigned,
                            to the given chemical
        """
        if value := self.standard_species_enthalpy.get(label):
            return value
        value = self.random_species_enthalpy()   # Assign a new value
        self.standard_species_enthalpy[label] = value
        #print(f"Standard species enthalpy of {value} assigned to `{label}`")
        return value



    def random_reaction_enthalpy(self, reactants :[str], products :[str]):
        """
        Attempt to provide a plausible regime for non-enzymatic small-molecule aqueous reactions,
        and thermodynamically consistent with all the previous random reactions added so far.

        It aims to randomly generate reaction ΔH values that are normally-distributed, with a mean of zero,
        and a standard deviation of roughly 70-90 kJ/mol

        With σ ≈ 70 kJ/mol for ΔH:
            ~68% of reactions fall within ±70 kJ/mol    (one standard deviation from the mean)
            ~95% within ±140 kJ/mol                     (2 standard deviations from the mean)

        That is the plausible range for non-enzymatic small-molecule chemistry.  [TODO: FURTHER VERIFY]

        So statistically:
            ✔ You’ll get many moderate reactions
            ✔ Some strongly exothermic/endothermic ones

        For a given elementary reaction, specified by its reactants and products,
        compute the Sum_i[v_i * H0_i] ,
        where i ranges over all the species in the reaction,
        H0_i is the standard enthalpy of that species,
        and v_i is the stoichiometric coefficient (negative for reactants,
        and positive for products)

        :param reactants:   A list of the labels of the reactants
        :param products:    A list of the labels of the reaction products
        :return:            A random value for the change in enthalpy of the given reaction,
                                normally-distributed with mean 0 and σ ≈ 70 kJ/mol,
                                thermodynamically consistent with all the previous random reactions added so far
        """
        reaction_enthalpy = 0

        if len(reactants) == 1:
            h = self.get_species_enthalpy(reactants[0])
            reaction_enthalpy -= h
        elif reactants[0] == reactants[1]:
            h = self.get_species_enthalpy(reactants[0])
            reaction_enthalpy -= 2 * h
        else:
            h_1 = self.get_species_enthalpy(reactants[0])
            h_2 = self.get_species_enthalpy(reactants[1])
            reaction_enthalpy = reaction_enthalpy - h_1 - h_2


        if len(products) == 1:
            h = self.get_species_enthalpy(products[0])
            reaction_enthalpy += h
        elif products[0] == products[1]:
            h = self.get_species_enthalpy(products[0])
            reaction_enthalpy += 2 * h
        else:
            h_1 = self.get_species_enthalpy(products[0])
            h_2 = self.get_species_enthalpy(products[1])
            reaction_enthalpy = reaction_enthalpy + h_1 + h_2


        return reaction_enthalpy



    def random_reaction_entropy(self, reaction_type :str) -> float:
        """
        Attempt to provide plausible random entropy values for non-enzymatic small-molecule aqueous reactions.

        Current starting point (to further verify):

        For decomposition reactions:
            ΔS ~ Normal with mean=40 and σ=20 ,  J/(mol·K)
            Note that 95% of the values will fall between 0 and 80 (i.e within 2σ of the mean)
            Further clipped to within the range [0, 100]

        For synthesis reactions:
            ΔS ~ Normal with mean=-40 and σ=20 ,  J/(mol·K)
            Note that 95% of the values will fall between -80 and 0 (i.e within 2σ of the mean)
            Further clipped to within the range [-100, 0]

        At biological temperature T ≈ 300K, our distribution centers of 40 J/mol·K will correspond to entropy values of:
            300 * 40 J/mol = 12 kJ/mol

        :param reaction_type:One of the following reaction types:
                                "ReactionSynthesis", "ReactionDecomposition"
        :return:            A random value for the change in entropy of the given reaction
        """
        if reaction_type == "ReactionDecomposition":
            entropy = self.rng.normal(loc=40, scale=20)    # Mean and SD
            if entropy < 0:
                return 0.
            if entropy > 100:
                return 100
        elif reaction_type == "ReactionSynthesis":
            entropy = self.rng.normal(loc=-40, scale=20)    # Mean and SD
            if entropy < -100:
                return -100.
            if entropy > 0:
                return 0
        else:
            raise Exception("random_reaction_entropy(): only `ReactionDecomposition` and `ReactionSynthesis` are currently allowed")

        return entropy



# Classes ReactionKinetics and VariableTimeSteps:

import math
import cmath
import numpy as np
import plotly.graph_objects as pgo
from life123.numerical import Numerical
from life123.visualization.plotly_helper import PlotlyHelper



class ReactionKinetics:
    """
    Static methods about reactions kinetics

    For background, see https://life123.science/reactions
    """



    @staticmethod
    def half_time_unimolecular_irreversible(kF) -> float:
        """
        Return the time taken for the reactant concentration in an irreversible unimolecular reaction
        to decrease by half (halfway to the asymptotic state, which happens to be zero)

        :param kF:  Forward reaction rate constant
        :return:    Reaction time taken for reactant concentration to decrease by half
        """
        return 0.6931471805599 / kF     # ln 2 / kF



    @staticmethod
    def half_time_relaxation_unimolecular_reversible(kF, kR) -> float:
        """
        Return the time taken for the reactant concentration in a reversible unimolecular reaction
        to decrease halfway to its asymptotic equilibrium state

        :param kF:  Forward reaction rate constant
        :param kR:  Reverse reaction rate constant
        :return:    Reaction time taken for reactant concentration
                        to decrease halfway to its asymptotic equilibrium state
        """
        # TODO: maybe simply combine with half_time_unimolecular_irreversible, with kR=0
        return 0.6931471805599 / (kF + kR)      # ln 2 / (kF + kR)


    @staticmethod
    def relaxation_time_unimolecular_reversible(kF, kR) -> float:
        """
        This is the same as the half-time of relaxation, within a factor of ln 2

        :param kF:  Forward reaction rate constant
        :param kR:  Reverse reaction rate constant
        :return:    A quantity named "relaxation time"
        """
        # TODO: is this method needed?
        return 1. / (kF + kR)



    @staticmethod
    def half_time_to_equilibrium_irreversible_synthesis(kF, A0, B0) -> float:
        """
        Return the time taken for the reactant concentration
        in a synthesis reaction A + B -> P
        to decrease halfway to their asymptotic equilibrium state.
        The time taken for 50% of the limiting reactant to be consumed

        :param kF:  Forward reaction rate constant
        :return:    Reaction time taken for reactant concentration
                        to decrease halfway to its asymptotic equilibrium state
        """
        #TODO: test
        #equil_concs = ReactionKinetics.compute_equilibrium_conc_first_order(kF=kF, kR=kR, a=1, A0=A0, b=1, B0=B0, p=1, P0=P0)

        delta_conc = A0 - B0
        if np.allclose(delta_conc, 0):
            return 1 / (kF * A0)

        if delta_conc > 0:      # When A0 > B0
            return math.log((2 * A0 - B0) / A0) / (kF * delta_conc)
        else:
            return - math.log((2 * B0 - A0) / B0) / (kF * delta_conc)   #TODO: test



    @staticmethod
    def exact_advance_unimolecular_irreversible(kF, A0, P0, t, incremental=False) -> float:
        """
        Exactly advance the concentrations
        in the ir-reversible 1st Order Reaction A -> P,
        from time 0 to time t,
        with the specified parameters.

        :param kF:  Forward reaction rate constant (the reverse one is taken to be zero)
        :param A0:  Initial concentration of the reactant A
        :param P0:  Initial concentration of the product P
        :param t:   The end time of the reaction that started at time zero
        :param incremental: [OPTIONAL] If True, the change in concentration is returned,
                                rather than the final ones.  Default: False

        :return:    The concentrations of P at time t (if `incremental` is False)
                        or its concentration change during the time interval (if `incremental` is True)
        """

        # Formula is:  A(t) = A0 Exp(-kF t)

        if incremental:
            A_incr = A0 * (np.exp(-kF * t) - 1)
            P_incr = - A_incr   # From mass conservation
            return P_incr


        TOT = A0 + P0
        A_t = A0 * np.exp(-kF * t)
        P_t = TOT - A_t     # From mass conservation

        return P_t



    @staticmethod
    def exact_advance_unimolecular_reversible(kF, kR, A0, P0, t, incremental=False) -> float:
        """
        Exactly advance the concentrations
        in the reversible elementary Reaction A <-> P,
        from time 0 to time t,
        with the specified parameters.

        :param kF:  Forward reaction rate constant
        :param kR:  Reverse reaction rate constant
        :param A0:  Initial concentration of the reactant A
        :param P0:  Initial concentration of the product P
        :param t:   The end time of the reaction that started at time zero
        :param incremental: [OPTIONAL] If True, the changes in concentrations is returned,
                                rather than the final one.  Default: False

        :return:    The concentrations of P at time t (if `incremental` is False)
                        or its concentration change during the time interval (if `incremental` is True)
        """
        TOT = A0 + P0

        # Formula is:  A(t) = (A0 - [(kR TOT) / (kF + kR)]) Exp[-(kF + kR) t] + kR TOT / (kF + kR)
        # In python:          (A0 - (kR * TOT) / (kF + kR)) * math.exp(-(kF + kR) * t) + kR * TOT / (kF + kR)

        sum_rates = kF + kR
        ratio = (kR * TOT) / sum_rates

        A_t = (A0 - ratio) * math.exp(-sum_rates * t) + ratio
        P_t = TOT - A_t     # From mass conservation

        if incremental:
            return P_t - P0

        return P_t



    @staticmethod
    def approx_solution_synthesis_rxn(kF, kR, A0, B0, P0, t :float|np.ndarray) -> float|np.ndarray:
        """
        Return the APPROXIMATE analytical solution, by way of exponentials,
        of the reversible 2nd Order Reaction A + B <-> P,
        with the specified parameters, sampled at the given time(s).

        This approximation is relatively coarse; in the pytests, errors of up to about 2% were seen,
        with the worst errors in the mid-portion between the starting state and the final equilibrium

        :param kF:  Forward reaction rate constant (cannot be zero)
        :param kR:  Reverse reaction rate constant
        :param A0:  Initial concentration of the 1st reactant A
        :param B0:  Initial concentration of the 2nd reactant B
        :param P0:  Initial concentration of the product P
        :param t:   The end time of the reaction that started at time zero
                        OR a Numpy array with the desired times at which the solutions are desired

        :return:    A concentration value, or Numpy arrays with the concentrations,
                        of P at the time - or times - given by the argument `t`
        """
        assert not np.allclose(kF, 0), \
            "approx_solution_synthesis_rxn(): this approximation cannot be used when kF is zero"

        # Calculate the equilibrium concentrations
        K_inv = kR / kF     # Inverse of the equilibrium constant.  Note that we're assuming kF isn't zero

        # The following derive from solving the quadratic equation  (A0 - m) * (B0 - m) / (C0 + m) == K_inv , for m
        # where m is the Product concentration change ("moles/liter of forward reaction")
        TOT_reactants = A0+B0
        r = (K_inv**2 + TOT_reactants**2) / 4 + (K_inv * TOT_reactants / 2) + P0 * K_inv - A0 * B0
        m = (K_inv + TOT_reactants)/2  - math.sqrt(r)         # Product concentration change

        # The reactants get consumed by m, while the product increases by m
        A_eq = A0 - m
        B_eq = B0 - m

        #print(f"\nProduct concentration change: {m} | A_eq: {A_eq}  |  B_eq: {B_eq}")

        l = (kF + kR) * (A_eq + B_eq)  # Relaxation rate constant λ

        AC_tot = A0 + P0        # Quantity conserved thru the rxn, from the stoichiometry

        A_arr = A_eq + (A0 - A_eq) * np.exp(-l * t)     # Approximate analytical solution
        P_arr = AC_tot - A_arr

        return P_arr



    @staticmethod
    def approx_solution_synthesis_rxn_ALT(kF, kR, A0, B0, P0, t, incremental) -> float:
        """
        Provided by ChatGPT.  Not fully tested.

        Closed-form solution for:
        dp/dt = kf (a0 - p + p0)(b0 - p + p0) - kr p
        with p(0) = p0

        Parameters
        ----------
        t : float or array_like
            Time(s) at which to evaluate p(t)
        P0, kF, kR, A0, B0 : float
            Reaction parameters

        Returns
        -------
        p : float or ndarray
            Value(s) of p(t)
        """

        # Coefficients of Riccati equation
        alpha = kF
        beta  = kF * (A0 + B0 + 2 * P0) + kR
        gamma = kF * (A0 + P0) * (B0 + P0)

        # Discriminant
        discr = beta**2 - 4*alpha*gamma
        if np.any(discr < 0):
            raise ValueError("Negative discriminant: parameters give complex roots.")

        sqrt_disc = np.sqrt(discr)

        # Equilibria
        p_plus  = (beta + sqrt_disc) / (2*alpha)
        p_minus = (beta - sqrt_disc) / (2*alpha)

        # Relaxation rate
        lam = alpha * (p_plus - p_minus)

        #t = np.asarray(t, dtype=float)

        exp_term = np.exp(-lam * t)

        numerator = (
                p_plus * (P0 - p_minus)
                - p_minus * (P0 - p_plus) * exp_term
        )

        denominator = (
                (P0 - p_minus)
                - (P0 - p_plus) * exp_term
        )

        P_t = numerator / denominator

        if incremental:
            return P_t - P0

        return P_t



    @staticmethod
    def exact_advance_synthesis_irreversible(kF, A0, B0, P0, t, incremental=False) -> float:
        """
        Exactly advance the concentrations
        in the irreversible elementary Reaction A + B -> P,
        from time 0 to time t,
        with the specified parameters.

        :param kF:  Forward reaction rate constant
        :param A0:  Initial concentration of the 1st reactant A
        :param B0:  Initial concentration of the 2nd reactant B
        :param P0:  Initial concentration of the product P
        :param t:   The end time of the reaction that started at time zero
        :param incremental: [OPTIONAL] If True, the changes in concentrations are returned,
                                rather than the final ones.  Default: False
        :return:    The concentration of P at time t (if `incremental` is False)
                        or its concentration change during the time interval (if `incremental` is True)
        """
        # Adapted from "Athel Cornish-Bowden, Fundamentals of Enzyme Kinetics, 4th edn", section 1.2.3
        # Note: the above author assumes P0 = 0, and overlooks the case A0 = B0 (which requires special treatment)
        '''
        Start with the ODE:
            d P(t) / dt = kF * A(t) * B(t) = kF * (A0 - (P(t)-P0)) * (B0 - (P(t)-P0))
                     
        Special case - when A0 = B0, the ODE becomes:
            d P(t) / dt = kF * A(t) * A(t) = kF * ( A0 - (P(t)-P0) )**2
            which can, for example, be plugged into Mathematica as:
            DSolve[{ p'[t] == kF (A0 + P0 - p[t])^2 , p[0] == P0 }, p[t], t]
        '''
        alpha =  P0 + A0

        if np.allclose(A0, B0):
            # When A0 = B0, the ODE has a special solution:
            #   P(t) = (P0 + alpha A0 kF t) / (1 + A0 kF t)
            prod = A0 * kF * t
            P_t = (P0 + alpha * prod) / (1 + prod)  # Concentration of `P` at time t

        else:
            # When A0 != B0, the solution P(t) of the ODE must satisfy:
            #   (A0 * (P0+B0 - P(t))) / (B0 * (P0+A0 - P(t))) = e ^ (B0-A0)*kF*t
            #   [see "Athel Cornish-Bowden, Fundamentals of Enzyme Kinetics, 4th edn", section 1.2.3]
            #   which gets solved for P(t) below
            #   Note: when A0 = B0, the above equation doesn't hold (at, at any rate, turns into a trivial 1 = 1)
            beta =   P0 + B0
            try:
                # Evaluate the exponential on the right-hand size
                eps = math.exp((B0-A0)* kF * t)
            except OverflowError:
                # Overflow occurs if B0 > A0 and t is very large;
                # in such a case, `A` is the limiting reagent, and fully converts to `P`
                if incremental:
                    return A0
                return alpha    # P0 + A0

            ''' 
            With the above variable assignments, and writing P for P(t), the earlier equation can be re-stated as:
                (A0 * (beta - P)) / (B0 * (alpha - P)) = eps 
                           
            i.e.:
                (A0 * (beta - P)) = eps * (B0 * (alpha - P))
                A0 * beta - A0 * P = eps * (B0 * alpha - B0 * P)
                A0 * beta - A0 * P = B0 * alpha * eps - B0 * P * eps
                B0 * P * eps - A0 * P = B0 * alpha * eps - A0 * beta
                (B0 * eps - A0) * P   = B0 * alpha * eps - A0 * beta
            '''
            #print(f"alpha: {alpha} , beta: {beta} , eps: {eps}")
            P_t = (B0 * alpha * eps - A0 * beta) / (B0 * eps - A0)    # Concentration of `P` at time t


        if incremental:
            return P_t-P0

        return P_t



    @staticmethod
    def exact_advance_synthesis_reversible(kF, kR, A0, B0, P0, t, incremental=False) -> float:
        """
        Exactly advance the concentrations
        in the reversible elementary Reaction A + B <-> P,
        from time 0 to time t,
        with the specified parameters.

        :param kF:  Forward reaction rate constant
        :param kR:  Reverse reaction rate constant
        :param A0:  Initial concentration of the 1st reactant A
        :param B0:  Initial concentration of the 2nd reactant B
        :param P0:  Initial concentration of the product P
        :param t:   The end time of the reaction that started at time zero
        :param incremental: [OPTIONAL] If True, the changes in concentrations are returned,
                                rather than the final ones.  Default: False

        :return:   The concentration of P at time t (if `incremental` is False)
                        or its concentration change during the time interval (if `incremental` is True)
        """
        if np.allclose(kR, 0):
            '''
            Note - the case when KR=0 and A0=B0 cannot be passed to the general solver below,
                   because in such as case:
                        T = A0 + P0     # A collapse of the general cases of AP_tot and BP_tot
                        alpha = kF
                        beta = 2 * kF * T
                        gamma = kF * T**2 
            In this scenario:
                - beta**2 + 4 * alpha * gamma = - [2 kF T]**2 + 4 kF * (kF * T**2) 
                                              = - 4 kF**2 T**2 + 4 kF**2 T**2 
                                              = 0  
            and the general solution cannot be used, because in it we divide by the square root of that term!       
            '''
            #print("****** Switching to IR-reversible version")
            return  ReactionKinetics.exact_advance_synthesis_irreversible(kF=kF, A0=A0, B0=B0, P0=P0, t=t, incremental=incremental)


        AP_tot = A0 + P0        # Quantity conserved thru the rxn, from the stoichiometry
        BP_tot = B0 + P0        # Quantity conserved thru the rxn, from the stoichiometry

        alpha = kF
        beta = kF * (AP_tot + BP_tot) + kR
        gamma = kF * AP_tot * BP_tot
        '''
        The ODE to solve for P(t) is:
            P'(t) = kF A(t) B(t) - kR P(t)      # Forward rate minus reverse rate
            
        The stoichiometry dictates that the drop in each reagent equals the increase in the product:
            A(t) = A0 - (P(t) - P0)
            B(t) = B0 - (P(t) - P0)
            
        which can be re-arranged to:
            P'(t) = kF [A0 - P(t) + P0] [B0 - P(t) + P0] - kR P(t)
            P'(t) = kF [AC_tot - P(t)] [BC_tot - P(t)] - kR P(t)
            P'(t) = kF [AC_tot * BC_tot - AC_tot * P(t) - P(t) * BC_tot + P(t)**2] - kR P(t)
            P'(t) = kF * P(t)**2 
                    - kF * AC_tot * P(t) - kF * BC_tot * P(t) - kR P(t) 
                    + kF * AC_tot * BC_tot 
            P'(t) = kF * P(t)**2 
                    - (kF * AC_tot + kF * BC_tot + kR) * P(t) 
                    +  kF * AC_tot * BC_tot 
            P'(t) = alpha [P(t)]**2 - beta P(t) + gamma
        
        The following is based on the answer by Mathematica v.7 to:
            DSolve[{p'[t] == alpha (p[t])**2 - beta p[t] + gamma, p[0] == P0}, p[t], t]
            Simplify[%]
        '''
        term = - beta**2 + 4 * alpha * gamma
        #print(f"term = {term}")

        if np.allclose(term, 0):
            # The term = 0 must be intercepted because we later divide by the square root of that term
            # TODO: can this scenario ever occur when kR > 0  ?
            raise Exception("exact_advance_synthesis_reversible(): No provision for this case yet, when beta**2 = 4 * alpha * gamma")
        else:
            sqrt_term = cmath.sqrt(term)            # Note: `term` could be negative; hence, the complex square root

            arctan_arg = (beta - 2 * alpha * P0) / sqrt_term    # This will be the argument to pass to the arctan() function, below
            arctan_value = np.arctan(arctan_arg)
            #print(f"arctan_value = {arctan_value}")

            # Note that every variable so far is a constant, NOT dependent on time

            tan_arg = (0.5 * sqrt_term * t) - arctan_value     # This will be the argument to pass to the tan() function, below

            result = (beta + sqrt_term * np.tan(tan_arg) ) / (2*alpha)

            P_t = result.real             # Drop the imaginary components (which ought to be very close to zero)


        if incremental:
            return P_t - P0

        return P_t



    @staticmethod
    def estimate_rate_constants_simple(t :np.ndarray,
                                       A_conc :np.ndarray, B_conc :np.ndarray,
                                       reactant_name="Reactant", product_name="Product"):
        """
        Estimate the rate constants for a 1-st order reaction of the type A <-> B,
        given time evolution of [A] and [B] on a grid of time points (the points don't need to be equally spaced),
        and create a plot to show the fit

        IMPORTANT : This is for reactions with a 1:1 stoichiometry between the given reactant and product

        :param t:               A numpy array of time grid points at which the other functions are specified
                                    (the points do not need to be equally-spaced)
        :param A_conc:          A numpy array of the concentrations of the reactant, at the times in the array t
        :param B_conc:          A numpy array of the concentrations of the product, at the times in the array t
        :param reactant_name:   [OPTIONAL] The name of the reactant (for display purposes)
        :param product_name:    [OPTIONAL] The name of the product (for display purposes)
        :return:                A plotly "Figure" object.  The estimated rate constants are printed out
        """
        total_conc_arr = A_conc + B_conc
        total_conc = np.median(total_conc_arr)    # TODO: give warning or abort if there's too much variance
        sd = np.std(total_conc_arr)

        print(f"Reaction {reactant_name} <-> {product_name}")
        print(f"Total REACTANT + PRODUCT has a median of {total_conc:,.4g}, "
              f"\n    with standard deviation {sd:,.4g} (ideally should be zero)")


        # The rate of change of reactant concentration with time
        A_prime = np.gradient(A_conc, t, edge_order=2)
        # The rate of change of product concentration with time
        B_prime = np.gradient(B_conc, t, edge_order=2)

        median_sum_derivs = np.median(A_prime + B_prime)
        print(f"The sum of the time derivatives of the reactant and the product "
              f"\n    has a median of {median_sum_derivs:,.4g} (ideally should be zero)")


        # Do a least-square fit
        kF, kR = Numerical.two_vector_least_square(V = A_conc, W = -B_conc, Y = B_prime)

        print(f"Least square fit to model as elementary reaction: {product_name}'(t) = kF * {reactant_name}(t) - kR * {product_name}(t)")

        # Plot both Y and its least-square fit, as functions of t
        fig_main = PlotlyHelper.plot_curves(x=t, y=[B_prime , kF * A_conc - kR * B_conc],
                                       title=f"d/dt {product_name}(t) and its least-square fit",
                                       x_label="t", y_label=f"d/dt  {product_name}(t)",
                                       legend_title="Curves",
                                       curve_labels=[f"d/dt {product_name}(t): exact", f"d/dt {product_name}(t): least-square fit"],
                                       colors=['green', 'red'])

        '''
        # Plot both Y and its least-square fit, as functions of X
        fig = PlotlyHelper.plot_curves(x=A_conc, y=[B_prime , kF * A_conc - kR * B_conc],
                                       title=f"d/dt {product_name}(t) as a function of {reactant_name}(t), alongside its least-square fit",
                                       x_label=f"{reactant_name}(t)", y_label=f"{product_name}'(t)",
                                       curve_labels=[f"{product_name}'(t)", "Linear Fit"], legend_title="Curve vs Fit:",
                                       colors=['green', 'red'])
        '''

        fig_side = PlotlyHelper.plot_curves(x=A_conc, y=B_prime,
                                         title=f"d/dt {product_name}(t) as a function of {reactant_name}(t)",
                                         x_label=f"{reactant_name}(t)", y_label=f"d/dt  {product_name}(t)",
                                         colors="purple")

        print(f"\n-> ESTIMATED RATE CONSTANTS: kF = {kF:,.4g} , kR = {kR:,.4g}")

        return PlotlyHelper.combine_in_vertical_grid(fig1=fig_main, fig2=fig_side,
                                                     title1=f"d/dt {product_name}(t) and its least-square fit",
                                                     title2=f"d/dt {product_name}(t) as a function of {reactant_name}(t)",
                                                     title_combined="LEAST-SQUARE FIT ANALYSIS:")



    @staticmethod
    def estimate_rate_constants_synthesis(t :np.ndarray,
                                          A_conc :np.ndarray, B_conc :np.ndarray, C_conc :np.ndarray,
                                          reactants :[str, str], product :str) -> pgo.Figure:
        """
        Estimate the rate constants for a 1-st order association (synthesis) reaction of the type A + B <-> C,
        given time evolution of [A], [B] and [C] on a grid of time points (don't need to be equally spaced)

        IMPORTANT : This is for reactions with a 1:1:1 stoichiometry

        :param t:           A numpy array of time grid points where the other functions are specified
        :param A_conc:      A numpy array of the concentrations of the reactant, at the times in the array t
        :param B_conc:
        :param C_conc:
        :param reactants:   [OPTIONAL] A list with the names of the 2 reactants, in order (for display purposes)
        :param product:     [OPTIONAL] The name of the product (for display purposes)
        :return:            A plotly "Figure" object.  The estimated rate constants are printed out
        """
        # The rate of change of [product] with time
        Deriv_C = np.gradient(C_conc, t, edge_order=2)

        # Do a least-square fit
        kF, kR = Numerical.two_vector_least_square(V = A_conc * B_conc, W = - C_conc, Y = Deriv_C)

        print(f"Least square fit to {product}'(t) = kF * {reactants[0]}(t) * {reactants[1]}(t) + kR * (- {product}(t) )")

        # Plot both Y and its least-square fit, as functions of X
        fig = PlotlyHelper.plot_curves(x=A_conc, y=[Deriv_C , kF * A_conc * B_conc - kR * C_conc],
                                       title=f"d/dt {product}(t) as a function of {reactants[0]}(t), alongside its least-square fit",
                                       x_label=f"{reactants[0]}(t)", y_label=f"{product}'(t)",
                                       curve_labels=[f"{product}'(t)", "Linear Fit"], legend_title="Curve vs Fit:",
                                       colors=['green', 'red'])

        print(f"\n-> ESTIMATED RATE CONSTANTS: kF = {kF:,.4g} , kR = {kR:,.4g}")

        return fig



    @staticmethod
    def compute_rate_elementary(reactants :[str], products :[str],
                                kF :float, kR :float, reversible :bool,
                                conc_dict :dict) -> float:
        """
        Given a SINGLE elementary reaction, in 1st order to all its chemical species,
        and the specified concentrations of chemicals,
        compute its initial reaction's "rate" (aka "velocity"),
        i.e. its "forward rate" minus its "reverse rate",
        at the start of the time step.

        Warning: even though arbitrary lists of reactants and products are accepted as arguments,
                 this kinetic model will generally only hold for unimolecular reactions A <-> P,
                 synthesis reactions A + B <-> P, and decomposition reactions A <-> B + C

        :param reactants:   List of labels of the reactants
        :param products:    List of labels of the products of the reaction
        :param kF:          Forward reaction rate
        :param kR:          Reverse reaction rate; ignored if irreversible
        :param reversible:  True if the reaction is reversible; False otherwise
        :param conc_dict:   A dict mapping chemical labels to their concentrations,
                                for all the chemicals involved in the given reaction
                                EXAMPLE:  {"B": 1.5, "F": 31.6, "D": 19.9}

        :return:            The differences between the reaction's forward and reverse rates
        """
        issue_warning = False
        if len(reactants) > 2 or len(products) > 2:
            issue_warning = True

        if len(reactants) == 2 and len(products) > 1:
            issue_warning = True

        if len(products) == 2 and len(reactants) > 1:
            issue_warning = True

        if issue_warning:
            print("compute_rate_elementary(): WARNING - using 1st order kinetic modeling "
                  "for a reaction that's probably not elementary")

        forward_rate = kF        # The initial multiplicative factor
        for r in reactants:
            # Process all the reactants
            conc = conc_dict[r]
            forward_rate *= conc

        if not reversible:
            return forward_rate

        reverse_rate = kR        # The initial multiplicative factor
        for p in products:
            # Process all the reaction products
            conc = conc_dict[p]
            reverse_rate *= conc

        return forward_rate - reverse_rate



    @staticmethod
    def compute_rate_mass_action_kinetics(reactant_terms :[(int, str)], product_terms :[(int, str)],
                                       kF :float, kR :float,
                                       conc_dict :dict) -> float:
        """
        Given a SINGLE arbitrary complex reaction,
        and the specified concentrations of its chemicals,
        compute its initial reaction "rate" (aka "velocity"),
        i.e. its "forward rate" minus its "reverse rate",
        at the start of the time step,
        when the reaction kinetics can be modeled as if it were an elementary reaction.

        This function is largely a convenient default function for testing,
        in scenarios (and hypothetical scenarios) where the the reaction's kinetics
        follow the familiar "Rate Laws",
        with the order of the reaction with respect to its reactant and products
        equals their respective stoichiometric coefficients -
        in other words an elementary reaction or one (hypothetically) modeled as such.

        For example, if the reaction is aA + bB <-> pP + qQ,
        then this function returns:  kF [P]^p [Q]^q - kR [A]^a [B]^b

        Warning: generally speaking, this is NOT a valid kinetic modeling
        of any reaction that isn't elementary

        :param reactant_terms:  A list of pairs (stoichiometry coefficient , label) for the reactants
        :param product_terms:   A list of pairs (stoichiometry coefficient , label) for the products
        :param kF:              Forward reaction rate
        :param kR:              Reverse reaction rate; zero if the reaction is irreversible

        :param conc_dict:       A dict mapping chemical labels to their concentrations,
                                    for all the chemicals involved in this reaction
                                    EXAMPLE:  {"B": 1.5, "F": 31.6, "D": 19.9}

        :return:            The differences between the reaction's forward and reverse rates
        """
        forward_rate = kF        # The initial multiplicative factor
        for order, reactant_label in reactant_terms:     # The stoichiometry coeff. of each reactant is taken to be its reaction order
            conc = conc_dict.get(reactant_label)
            assert conc is not None, \
                f"compute_rate_mass_action_kinetics(): missing concentration value for reactant chemical with label `{reactant_label}`"
            forward_rate *= conc ** order       # Raise to power


        if kR == 0:
            return forward_rate                 # If there's no reverse reaction (i.e., if reaction is irreversible)


        reverse_rate = kR        # The initial multiplicative factor
        for order, product_label in product_terms:      # The stoichiometry coeff. of each product is taken to be its reaction order
            conc = conc_dict.get(product_label)
            assert conc is not None, \
                f"compute_rate_mass_action_kinetics(): missing concentration value for product chemical `{product_label}`"
            reverse_rate *= conc ** order       # Raise to power

        return forward_rate - reverse_rate


    @staticmethod
    def compute_rate_first_order(reactant_terms :[(int, str)], product_terms :[(int, str)],
                                       kF :float, kR :float,
                                       conc_dict :dict) -> float:
        """
        If the reactions isn't elementary, this is a HYPOTHETICAL scenario (mostly for testing and analysis)
        where the reaction is first order in all reactants and products

        :param reactant_terms:
        :param product_terms:
        :param kF:
        :param kR:
        :param conc_dict:
        :return:
        """
        # Pretend that the reaction is an elementary one
        reactants = [t[1] for t in reactant_terms]
        products  = [t[1] for t in product_terms]
        return ReactionKinetics.compute_rate_elementary(reactants=reactants, products=products, kF=kF, kR=kR, reversible=True, conc_dict=conc_dict)



    @staticmethod
    def compute_equilibrium_conc_first_order(kF, kR, a, A0, p, P0, b=0, B0=None, q=0, Q0=None) -> dict:
        """
        Determine the equilibrium concentrations that would be eventually reached
        by the chemicals participating in a reversible reaction of the form:
            a A + b B <-> p P + q Q
        whose kinetics are (hypothetically) FIRST ORDER in each of the chemicals
        (some of the terms may be missing),
        given the specified initial concentrations,
        IN THE ABSENCE of any other reaction.

        In other words, this applies to reactions that can be modeled kinetically as:
            vF = kF [A] [B]  ,  vR = kR [C] [D]    (some of the terms may be missing)
        where vF and vR are, respectively, the forward and reverse reaction velocities (rates).
        If the reaction isn't elementary, kF will be a composite "kF effective"; likewise for kR.

        CAUTION: since all reaction orders are expected to be 1,
                 if any of stoichiometry coefficients of the involved chemicals aren't 1,
                 then it's a scenario of "generalized kinetics"
                 (non-mass-action with respect to the stoichiometry)

        Note:  we're using "concentrations" instead of "chemical activities";
               concentrations approximate the activities of ideal dilute solutions

        :param kF:  The reaction's forward rate constant.
                        If the reaction isn't elementary (but can be modeled as being of 1st order in each of the chemicals),
                        then this value will be a composite "kF effective"
        :param kR:  The reaction's reverse rate constant.
                        If the reaction isn't elementary (but can be modeled as being of 1st order in each of the chemicals),
                        then this value will be a composite "kR effective"

        :param a:   The stoichiometry coefficient of chemical A in the reaction (typically 1)
        :param A0:  The initial concentration of chemical A
        :param b:   [OPTIONAL] The stoichiometry coefficient of chemical B in the reaction;
                        if not present, accept the default value of zero
        :param B0:  [OPTIONAL] The initial concentration of chemical B;
                        if not present, accept the default None
        :param p:   The stoichiometry coefficient of product chemical `P` in the reaction (typically 1)
        :param P0:  The initial concentration of the product chemical `P`
        :param q:   [OPTIONAL] The stoichiometry coefficient of chemical D in the reaction;
                        if not present, accept the default value of zero
        :param Q0:  [OPTIONAL] The initial concentration of chemical B;
                        if not present, accept the default value

        :return:    A dictionary of the equilibrium concentrations of the
                        chemicals involved in the specified reaction;
                        only the applicable entries will be present
                        EXAMPLES:
                            {'A': 24.0, 'B': 36.0, 'P': 1.8, 'Q': 0.3}
                            {'A': 35.0, 'P': 18.3}
        """
        '''
        For (hypothetical) reactions of the form aA + bB <-> cC + dD  
        that are FIRST-ORDER in all chemicals, and with the given initial condition,
        the equilibrium equation needs to equate the reaction's forward and reverse rates 
        after the reaction has advanced by m moles 
        (which consumes reagents in proportion to their stoichiometric coefficients, and 
        generates products similarly):
                  
            kF [(A0 - a*m) (B0 - b*m)] = kR [(C0 + c*m) (D0 + d*m)]     # Forward rate = Reverse rate
        
        where m (to be solved for) is the number of "moles/liter of forward reaction" 

        Generalization 1: 
        The above equilibrium equation can also be made to handle reaction terms that aren't present 
        (for example the "D" part in the reaction A + B <-> C), 
        by simply using 1 for the "initial concentration" and 0 for the "stoichiometry coefficient"
        of any missing term;
        such a trick will make its corresponding multiplicative term to become (1 + 0*m) ,
        i.e. identically equal to 1 , and thus have no effect on the solution of the equation.
        
        Generalization 2:
        The equilibrium equation can also be made to handle scenarios where 2 terms refer to the same chemical,
        such as in the reaction 2 A <-> C , with 2nd order with respect to A, 
        by setting B0 = A0, a=2, b=2,
        which will produce the multiplicative term that we need, namely:
            kF [(A0 - 2*m) **2]
        
        
        SOLUTION -
        Our equilibrium equation can be expanded into a standard quadratic form for the unknown m :
        
                alpha * m**2 + beta * m + gamma = 0
        
        (where alpha, beta and gamma are as computed below), and then solved for m.  
        In case of two solution for the quadratic, we'll pick the one that
        leads to physically-possible results (non-negative concentrations of all the chemicals.)
        '''
        assert kR is not None, \
            "compute_equilibrium_conc_first_order(): a value must be passed for argument `kR` (currently None)"

        assert kF is not None, \
            "compute_equilibrium_conc_first_order(): a value must be passed for argument `kF` (currently None)"

        if b == 0:
            assert B0 is None, \
                "compute_equilibrium_conc_first_order(): unexpected concentration value for a chemical not in reaction (B)"
            B0 = 1      # A trick to reduce to the general equation when b=0

        if q == 0:
            assert Q0 is None, \
                "compute_equilibrium_conc_first_order(): unexpected concentration value for a chemical not in reaction (D)"
            Q0 = 1      # A trick to reduce to the general equation when d=0

        alpha = (kF * a * b - kR * p * q)
        beta = -kF * (b*A0 + a*B0) -kR * (q * P0 + p * Q0)
        gamma = kF * A0 * B0 - kR * P0 * Q0

        #print("compute_equilibrium_conc_first_order() - alpha, beta, gamma : ", alpha, beta, gamma)

        if np.allclose(alpha, 0):
            # The quadratic reduces to the linear equation:  beta * m + gamma = 0
            m1 = -gamma / beta
            m2 = m1
        else:
            sqrt_discriminant = math.sqrt(beta**2 - 4 * alpha * gamma)
            m1 = (-beta - sqrt_discriminant) / (2 * alpha)
            m2 = (-beta + sqrt_discriminant) / (2 * alpha)
            #print("m1, m2 : ", m1, m2)

        m = m1  # Let's start with one of the 2 possible solutions of the quadratic equation

        # After m "moles of forward reaction", the concentration of the reactant "A"
        # in aA + bB <-> cC + dD gets reduced by a*m . Likewise for the other terms.
        # Reaction products get increased.  Values for missing terms will be meaningless
        std_result = {"A" : A0 - a*m, "B" : B0 - b*m, "P" : P0 + p * m, "Q" : Q0 + q * m}   # TENTATIVE values!

        if min(std_result.values()) < 0:    # If there's any negative value in the concentrations...
            # ...then repeat the computation using the other solution of the quadratic
            m = m2
            print("Using 2nd solution: m = ", m)
            std_result = {"A" : A0 - a*m, "B" : B0 - b*m, "C" : P0 + p * m, "D" : Q0 + q * m}

        # Eliminate entries that aren't applicable to the reaction
        if b == 0:
            del std_result["B"]
        if q == 0:
            del std_result["Q"]

        return std_result







####################################################################################################

class VariableTimeSteps:
    """
    Methods for managing variable time steps during reactions
    """


    def __init__(self):
        # ***  PARAMETERS FOR AUTOMATED ADAPTIVE TIME STEP SIZES  ***
        # Note: The "aborts" below are "elective" aborts - i.e. not aborts from hard errors (further below)
        #       The default values get packed into a "preset", specified by a name, and optionally passed when
        #       instantiating this object.
        #       Users in need to change these values post-instantiation will generally use use_adaptive_preset(), below;
        #       or, for more control, set_thresholds() and set_step_factors()
        self.thresholds = []    # A list of "rules"
                                # EXAMPLE:                  [
                                #                             {"norm": "norm_A", "low": 0.5, "high": 0.8, "abort": 1.44},
                                #                             {"norm": "norm_B", "low": 0.08, "high": 0.5, "abort": 1.5}
                                #                           ]
                                #       low:    The value below which the norm is considered to be good (and the variable steps are being made larger);
                                #                if None, it doesn't get used
                                #       high:   The value above which the norm is considered to be excessive (and the variable steps get reduced);
                                #               if None, it doesn't get used
                                #       abort:  The value above which the norm is considered to be dangerously large (and the last variable step gets
                                #               discarded and back-tracked with a smaller size)
                                #               if None, it doesn't get used

        self.step_factors = {}
                            # EXAMPLE: {"upshift": 1.2, "downshift": 0.5, "abort": 0.4, "error": 0.2}
                            # "upshift" must be > 1 ; all the other values must be < 1
                            # Generally, error <= abort <= downshift
                            # "Error" value: Factor by which to multiply the time step
                            #   in case of negative-concentration error from excessive step size
                            #   NOTE: this is from ERROR aborts,
                            #   not to be confused with general aborts based on reaching high threshold


        # Zero the number of times that each norm got involved in step-size decision
        self.norm_usage = {}
        self.reset_norm_usage_stats()




    def set_thresholds(self, norm :str, low=None, high=None, abort=None) -> None:
        """
        Create or update a rule based on the given norm
        (simulation parameters that affect the adaptive variable step sizes.)
        If no rule with the specified norm was previously set, it will be added.

        All None values in the arguments are ignored.

        :param norm:    The name of the "norm" (criterion) being used - either a previously-set one or not
        :param low:     The value below which the norm is considered to be good (and the variable steps are being made larger);
                            if None, it doesn't get changed
        :param high:    The value above which the norm is considered to be excessive (and the variable steps get reduced);
                            if None, it doesn't get changed
        :param abort:   The value above which the norm is considered to be dangerously large (and the last variable step gets
                            discarded and back-tracked with a smaller size) ;
                            if None, it doesn't get changed
        :return:        None
        """
        assert type(norm) == str and norm != "", \
            "set_thresholds(): the `norm` argument must be a non-empty string"

        if (abort is not None) and (high is not None):
            assert abort > high, \
                f"set_thresholds(): `abort` value ({abort}) must be > `high' value ({high})"

        if (high is not None) and (low is not None):
            assert high > low, \
                f"set_thresholds(): `high` value ({high}) must be > `low' value ({low})"

        if (abort is not None) and (low is not None):
            assert abort > low, \
                f"add_thresholds(): `abort` value ({high}) must be > `low' value ({low})"

        for i, t in enumerate(self.thresholds):
            if t.get("norm") == norm:
                # Found a rule using the requested norm

                t_original = t.copy()  # Create a backup copy in case of error

                if low is not None:
                    t["low"] = low

                if high is not None:
                    t["high"] = high

                if abort is not None:
                    t["abort"] = abort

                if len(t) == 1:
                    del self.thresholds[i]      # Completely eliminate this un-used norm

                low, high, abort = t.get("low"), t.get("high"), t.get("abort")

                if (abort is not None) and (high is not None):
                    if abort <= high:
                        self.thresholds[i] = t_original     # Restore original values
                        raise Exception(f"set_thresholds(): `abort` value ({abort}) must be > `high' value ({high})")

                if (high is not None) and (low is not None):
                    if high <= low:
                        self.thresholds[i] = t_original     # Restore original values
                        raise Exception(f"set_thresholds(): `high` value ({high}) must be > `low' value ({low})")

                if (abort is not None) and (low is not None):
                    if abort <= low:
                        self.thresholds[i] = t_original     # Restore original values
                        raise Exception(f"add_thresholds(): `abort` value ({high}) must be > `low' value ({low})")

                return

        # If we get here, it means that we're handling a norm
        # not currently present in the list self.thresholds
        new_t = {"norm": norm}
        if low is not None:
            new_t["low"] = low
        if high is not None:
            new_t["high"] = high
        if abort is not None:
            new_t["abort"] = abort

        if self.thresholds is None:
            self.thresholds = [new_t]
        else:
            self.thresholds.append(new_t)



    def delete_thresholds(self, norm :str, low=False, high=False, abort=False) -> None:
        """
        Delete one or more of the threshold values associated to a rule using the specified norm.
        If none of the threshold values remains in place, the whole rule gets eliminated altogether.
        Attempting to delete something not present, will raise an Exception

        :param norm:    The name of the "norm" (criterion) being used - for a previously-set rule
        :param low:     If True, this threshold value will get deleted
        :param high:    If True, this threshold value will get deleted
        :param abort:   If True, this threshold value will get deleted
        :return:        None
        """
        for i, t in enumerate(self.thresholds):
            if t.get("norm") == norm:
                # Found a rule using the requested norm

                if low:
                    del t["low"]

                if high:
                    del t["high"]

                if abort:
                    del t["abort"]

                if len(t) == 1:
                    del self.thresholds[i]      # Completely eliminate this un-used norm
                return

        # If we get here, it means that we're handling a norm
        # not currently present in the list self.thresholds
        raise Exception(f"delete_thresholds(): no norm named '{norm}' was found")



    def display_value_against_thresholds(self, all_norms) -> None:
        """

        :param all_norms:
        :return:            None
        """
        for rule in self.thresholds:
            value = all_norms.get(rule['norm'])
            print(self.display_value_against_thresholds_single_rule(rule, value))


    def display_value_against_thresholds_single_rule(self, rule :dict, value) -> str:
        """
        Examine how the specified value fits
        relatively to the 'low', 'high', and 'abort' stored in the given rule

        :param rule:    A dict that must contain the key 'norm',
                            and may contain the keys: 'low', 'high', and 'abort'
                            (referring to 3 increasingly-high thresholds)
        :param value:   Either None, or a number to compare to the thresholds
        :return:        A string that visually highlights the relative position of the value
                            relatively to the given thresholds
        """
        s = f"                   {rule['norm']} : "     # Name of the norm being used

        if value is None:
            return s + " (skipped; not needed)"

        # Extract the 3 thresholds (some might be missing)
        low = rule.get('low')
        high = rule.get('high')
        abort = rule.get('abort')

        if low is not None and value <= low:
            # The value is below the `low` threshold
            return f"{s}(VALUE {value:.5g}) | low {low} | high {high} | abort {abort}"

        if high is not None and value < high:
            # The value is between the `low` and `high` thresholds
            return f"{s}low {low} | (VALUE {value:.5g}) | high {high} | abort {abort}"

        if abort is not None and value < abort:
            # The value is above the `high` threshold
            return f"{s}low {low} | high {high} | (VALUE {value:.5g}) | abort {abort}"

        # If we get thus far, the value is above the `abort` threshold
        return f"{s}low {low} | high {high} | abort {abort} | (VALUE {value:.5g})"



    def set_step_factors(self, upshift=None, downshift=None, abort=None, error=None) -> None:
        """
        Over-ride current values for simulation parameters that affect the adaptive variable step sizes.
        Values not explicitly passed will remain the same.

        :param upshift:     Fraction by which to increase the variable step size
        :param downshift:   Fraction by which to decrease the variable step size
        :param abort:       Fraction by which to decrease the variable step size in cases of step re-do
        :param error:       Fraction by which to decrease the variable step size in cases of error
        :return:            None
        """
        if upshift is not None:
            assert upshift > 1, "set_step_factors(): `upshift` value must be > 1"
            self.step_factors["upshift"] = upshift

        if downshift is not None:
            assert 0 < downshift < 1, "set_step_factors(): `downshift` value must be a positive number < 1"
            self.step_factors["downshift"] = downshift

        if abort is not None:
            assert 0 < abort < 1, "set_step_factors(): `abort` value must be a positive number < 1"
            self.step_factors["abort"] = abort

        if error is not None:
            assert 0 < error < 1, "set_step_factors(): `error` value must be a positive number < 1"
            self.step_factors["error"] = error



    def show_adaptive_parameters(self) -> None:
        """
        Print out the current values for the adaptive time-step parameters

        :return:    None
        """
        print("Parameters used for the automated adaptive time step sizes -")
        print("    THRESHOLDS: ", self.thresholds)
        print("    STEP FACTORS: ", self.step_factors)



    def use_adaptive_preset(self, preset :str) -> None:
        """
        Lets the user choose a preset to use from now on, unless later explicitly changed,
        for use in ALL reaction simulations involving adaptive time steps.
        The preset will affect the degree to which the simulation will be "risk-taker" vs. "risk-averse" about
        taking larger steps.

        Note:   for more control, use set_thresholds() and set_step_factors()

                For example, using the "mid" preset is the same as issuing:
                    ReactionKinetics.set_thresholds(norm="norm_A", low=0.5, high=0.8, abort=1.44)
                    ReactionKinetics.set_thresholds(norm="norm_B", low=0.08, high=0.5, abort=1.5)
                    ReactionKinetics.set_step_factors(upshift=1.2, downshift=0.5, abort=0.4, error=0.25)

        :param preset:  String with one of the available preset names;
                            allowed values are (in generally-increasing speed):
                            'heavy_brakes', 'slower', 'slow', 'mid', 'fast'
        :return:        None
        """
        if preset == "heavy_brakes":   # It slams on the "brakes" hard in case of abort or errors
            self.thresholds = [{"norm": "norm_A", "low": 0.02, "high": 0.025, "abort": 0.03},
                               {"norm": "norm_B", "low": 0.05, "high": 1.0, "abort": 2.0}]
            self.step_factors = {"upshift": 1.6, "downshift": 0.15, "abort": 0.08, "error": 0.05}

        elif preset == "small_rel_change":
            self.thresholds = [{"norm": "norm_A", "low": 2., "high": 5., "abort": 10.},
                               {"norm": "norm_B", "low": 0.008, "high": 0.5, "abort": 2.0}]     # The "low" value of "norm_B" is very strict
            self.step_factors = {"upshift": 1.5, "downshift": 0.25, "abort": 0.25, "error": 0.2}

        elif preset == "slower":   # Very conservative about taking larger steps
            self.thresholds = [{"norm": "norm_A", "low": 0.2, "high": 0.5, "abort": 0.8},
                               {"norm": "norm_B", "low": 0.03, "high": 0.05, "abort": 0.5}]
            self.step_factors = {"upshift": 1.01, "downshift": 0.5, "abort": 0.1, "error": 0.1}

        elif preset == "slow":   # Conservative about taking larger steps
            self.thresholds = [{"norm": "norm_A", "low": 0.2, "high": 0.5, "abort": 0.8},
                               {"norm": "norm_B", "low": 0.05, "high": 0.4, "abort": 1.3}]
            self.step_factors = {"upshift": 1.1, "downshift": 0.3, "abort": 0.2, "error": 0.1}

        elif preset == "mid":     # A "middle-of-the road" heuristic: somewhat "conservative" but not overly so
            self.thresholds = [{"norm": "norm_A", "low": 0.5, "high": 0.8, "abort": 1.44},
                               {"norm": "norm_B", "low": 0.08, "high": 0.5, "abort": 1.5}]
            self.step_factors = {"upshift": 1.2, "downshift": 0.5, "abort": 0.4, "error": 0.25}

        elif preset == "fast":   # Less conservative (more "risk-taker") about taking larger steps
            self.thresholds = [{"norm": "norm_A", "low": 0.8, "high": 1.2, "abort": 1.7},
                               {"norm": "norm_B", "low": 0.15, "high": 0.8, "abort": 1.8}]
            self.step_factors = {"upshift": 1.5, "downshift": 0.8, "abort": 0.6, "error": 0.5}

        elif preset == "mid_inclusive": # A "middle-of-the road" heuristic that makes use of more norms
            self.thresholds = [{'norm': 'norm_A', 'low': 0.2, 'high': 0.8, 'abort': 1.44},
                               {'norm': 'norm_B', 'low': 0.08, 'high': 0.5, 'abort': 1.5},
                               {'norm': 'norm_C', 'low': 0.5, 'high': 1.2, 'abort': 1.6},
                               {'norm': 'norm_D', 'low': 1.3, 'high': 1.7, 'abort': 1.8}]
            self.step_factors = {'upshift': 1.1, 'downshift': 0.5, 'abort': 0.4, 'error': 0.25}

        elif preset == "mid_inclusive_slow":
            self.thresholds = [{'norm': 'norm_A', 'low': 0.15, 'high': 0.8, 'abort': 1.44},
                               {'norm': 'norm_B', 'low': 0.05, 'high': 0.5, 'abort': 1.5},
                               {'norm': 'norm_C', 'low': 0.5, 'high': 1.2, 'abort': 1.6},
                               {'norm': 'norm_D', 'low': 1.1, 'high': 1.7, 'abort': 1.8}]
            self.step_factors = {'upshift': 1.1, 'downshift': 0.5, 'abort': 0.4, 'error': 0.25}

        else:
            raise Exception(f"set_adaptive_parameters(): unknown value for the `preset` argument ({preset}); "
                            f"allowed values are 'heavy_brakes', 'slower', 'slow', 'mid', 'fast', 'mid_inclusive'")



    def adjust_timestep(self, n_chems: int, indexes_of_active_chemicals :[int],
                        delta_conc: np.array, baseline_conc=None, prev_conc=None,
                        ) -> dict:
        """
        Computes some measures of the change of concentrations, from the last step, in the context of the
        baseline initial concentrations of that same step, and the concentrations in the step before that.
        Based on the magnitude of the measures, propose a course of action about what to do for the next step.

        :param n_chems:         The total number of registered chemicals - exclusive of water and of macro-molecules
        :param indexes_of_active_chemicals: The ordered list (numerically sorted) of the INDEX numbers of all the chemicals
                                                involved in ANY of the registered reactions,
                                                but NOT counting chemicals that always appear in a catalytic role in all the reactions they
                                                participate in

        :param delta_conc:      A numpy array of changes in concentrations for the chemicals of interest,
                                    across a simulation time step (typically, the current step a run in progress)
        :param baseline_conc:   A numpy array of baseline concentration values for those same chemicals,
                                    prior to the above change, at the start of a simulation time step
        :param prev_conc:       A numpy array of concentration values for those same chemicals,
                                    in the step prior to the current one (i.e. an "archive" value)
        :return:                A dict:
                                    "action"           - String with the name of the computed recommended action:
                                                            either "low", "stay", "high" or "abort"
                                    "step_factor"      - A factor by which to multiply the time step at the next iteration round;
                                                            if no change is deemed necessary, 1
                                    "norms"            - A dict of all the computed norm name/values (any of the norms, except norm_A,
                                                            may be missing)
                                    "applicable_norms" - The name of the norm that triggered the decision; if all norms were involved,
                                                            it will be "ALL"
        """
        if baseline_conc is not None:
            assert len(baseline_conc) == len(delta_conc), \
                f"adjust_timestep(): the number of entries in the passed array `delta_conc` ({len(delta_conc)}) " \
                f"does not match the number of entries in the passed array `baseline_conc` ({len(baseline_conc)})"

        assert n_chems == len(delta_conc), \
            f"adjust_timestep(): the number of entries in the passed array `delta_conc` ({len(delta_conc)}) " \
            f"does not match the number of registered chemicals ({n_chems})"


        # If some chemicals are not dynamically involved in the reactions
        # (i.e. if they don't occur in any reaction, or occur as enzyme),
        # restrict our consideration to only the dynamically involved ones
        # CAUTION: the concept of "active chemical" might change in future versions, where only SOME of
        #          the reactions are simulated
        #if self.chem_data.number_of_active_chemicals() < n_chems:
        if len(indexes_of_active_chemicals) < n_chems:
            delta_conc = delta_conc[indexes_of_active_chemicals]
            #print(f"\nadjust_timestep(): restricting adaptive time step analysis to {n_chems} chemicals only; their delta_conc is {delta_conc}")
            if baseline_conc is not None:
                baseline_conc = baseline_conc[indexes_of_active_chemicals]
            if prev_conc is not None:
                prev_conc = prev_conc[indexes_of_active_chemicals]
            # Note: setting delta_conc, etc, only affects local variables, and won't mess up the arrays passed as arguments

        all_norms = {}

        all_small = True            # Provisional answer to the question: "do ALL the rule yield a 'low'?"
        # Any rule failing to yield a 'low' will flip this status

        high_seen_at = []           # List of rule names at which a "high" is encountered, if applicable

        for rule in self.thresholds:
            norm_name = rule["norm"]

            if norm_name == "norm_A":
                result = self.norm_A(delta_conc)
            elif norm_name == "norm_B":
                result = self.norm_B(baseline_conc, delta_conc)
            elif norm_name == "norm_C":
                result = self.norm_C(prev_conc, baseline_conc, delta_conc)
            else:
                result = self.norm_D(prev_conc, baseline_conc, delta_conc)

            all_norms[norm_name] = result

            if ("abort" in rule) and (result > rule["abort"]):
                # If any rules declares an abort, no need to proceed further: it's an abort
                #self.norm_usage[norm_name] += 1
                self.increase_norm_count(norm_name)
                return {"action": "abort", "step_factor": self.step_factors["abort"], "norms": all_norms, "applicable_norms": [norm_name]}

            if ("high" in rule) and (result > rule["high"]):
                # If any rules declares a "high", still need to consider the other rules - in case any of them over-rides
                # the "high" with an "abort"
                high_seen_at.append(norm_name)
                all_small = False           # No longer the case of all 'low` yields

            if all_small and ("low" in rule) and (result > rule["low"]):
                all_small = False           # No longer the case of all 'low` yields
        # END for


        if high_seen_at:
            for n in high_seen_at:
                #self.norm_usage[n] += 1
                self.increase_norm_count(n)
            return {"action": "high", "step_factor": self.step_factors["downshift"], "norms": all_norms, "applicable_norms": high_seen_at}


        if all_small:
            for i in self.norm_usage:
                # All the norms were used
                self.increase_norm_count(i)
                #self.norm_usage[i] += 1

            return {"action": "low", "step_factor": self.step_factors["upshift"], "norms": all_norms, "applicable_norms": "ALL"}


        # If we get thus far, none of the rules were found applicable
        return {"action": "stay", "step_factor": 1, "norms": all_norms, "applicable_norms": "ALL"}



    def relative_significance(self, value :float, baseline :float) -> str:
        """
        Estimate, in a loose categorical fashion, the magnitude of the quantity "value"
        in proportion to the quantity "baseline".
        Both are assumed non-negative (NOT checked.)
        Return one of:
            "S" ("Small" ; up to 1/2 the size)
            "C" ("Comparable" ; from 1/2 to double)
            "L" ("Large" ; over double the size)
        This method is meant for large-scale computations, and on purpose avoids doing divisions.

        NOT IN CURRENT ACTIVE USAGE (in former use for the discontinued substep implementation)

        :param value:
        :param baseline:
        :return:        An assessment of relative significance, as one of
                        "S" ("Small"), "C" ("Comparable"), "L" ("Large")
        """
        #TODO: a Numpy array version

        if value < baseline:
            if value + value < baseline:
                return "S"
            else:
                return "C"

        else:
            if baseline + baseline < value:
                return "L"
            else:
                return "C"





    #####################################################################################################

    '''                                         ~  NORMS  ~                                           '''

    def ________NORMS________(DIVIDER):
        pass         # Used to get a better structure view in IDEs such asPycharm
    #####################################################################################################


    def reset_norm_usage_stats(self):
        """
        Reset the count of the number of times that each norm got involved in step-size decision

        :return:    None
        """
        self.norm_usage = {"norm_A": 0, "norm_B": 0, "norm_C": 0, "norm_D": 0}



    def increase_norm_count(self, norm_name :str) -> None:
        """

        :param norm_name:
        :return:            None
        """
        assert norm_name in self.norm_usage, \
            f"increase_norm_count(): unknown norm named `{norm_name}`"

        self.norm_usage[norm_name] += 1



    def norm_A(self, delta_conc :np.array) -> float:
        """
        Return a measure of system change, based on the average concentration changes
        of ALL the specified chemicals across a time step, adjusted for the number of chemicals.
        A square-of-sums computation (the square of an L2 norm) is used.

        :param delta_conc:  A Numpy array with the concentration changes
                                of the chemicals of interest across a time step
        :return:            A measure of change in the concentrations across the simulation step
        """
        n_active_chemicals = len(delta_conc)

        assert n_active_chemicals > 0, \
            "norm_A(): zero-sized array was passed as argument"

        # The following are normalized by the number of chemicals
        #L2_rate = np.linalg.norm(delta_concentrations) / n_chems
        #L2_rate = np.sqrt(np.sum(delta_concentrations * delta_concentrations)) / n_chems
        #print("    L_inf norm:   ", np.linalg.norm(delta_concentrations, ord=np.inf) / delta_time)
        #print("    Adjusted L1 norm:   ", np.linalg.norm(delta_concentrations, ord=1) / n_chems)

        adjusted_L2_rate = np.sum(delta_conc * delta_conc) / (n_active_chemicals * n_active_chemicals)
        return adjusted_L2_rate



    def norm_B(self, baseline_conc: np.array, delta_conc: np.array) -> float:
        """
        Return a measure of system change, based on the max absolute relative concentration
        change of all the chemicals across a time step (based on an L infinity norm - but disregarding
        any baseline concentration that is very close to zero)

        :param baseline_conc:   A Numpy array with the concentration of the chemicals of interest
                                    at the start of a simulation time step
        :param delta_conc:      A Numpy array with the concentration changes
                                    of the chemicals of interest across a time step
        :return:                A measure of change in the concentrations across the simulation step
        """
        arr_size = len(baseline_conc)
        assert len(delta_conc) == arr_size, "norm_B(): mismatch in the sizes of the 2 passed array arguments"

        to_keep = ~ np.isclose(baseline_conc, 0)    # Element-wise negation; this will be an array of Booleans
        # with True for all the elements of baseline_conc that aren't too close to 0

        ratios = delta_conc[to_keep] / baseline_conc[to_keep]   # Using boolean indexing to only select some of the elements :
        # the non-zero denominators, and their corresponding numerators
        if len(ratios) == 0:
            return 0.

        return max(abs(ratios))



    def norm_C(self, prev_conc: np.array, baseline_conc: np.array, delta_conc: np.array) -> float:
        """
        Return a measure of system short-period oscillation; larger values might be heralding
        onset of simulation instability

        :param prev_conc:       A numpy array with the concentration of the chemicals of interest,
                                    in the step prior to the current one (i.e. an "archive" value)
        :param baseline_conc:   A Numpy array with the concentration of the chemicals of interest
                                    at the start of a simulation time step
        :param delta_conc:      A Numpy array with the concentration changes
                                    of the chemicals of interest across a time step
        :return:
        """
        if prev_conc is None:
            return 0            # Unable to compute a norm; 0 represents "perfect"

        D1 = baseline_conc - prev_conc
        D2 = delta_conc

        #print("****** D1: ", D1)
        #print("****** D2: ", D2)

        sign_flip = ((D1 >= 0) & (D2 < 0)) | ((D1 < 0) & (D2 >= 0))

        criterion_met = sign_flip & (abs(D2) > abs(D1)) & ~np.isclose(D1, 0) & (abs(D2) < 50 * abs(D1))
        #criterion_met = sign_flip & ~np.isclose(D1, 0) & (abs(D2) < 50 * abs(D1))
        # Note: values with D1 very close to zero are ignored
        #       likewise, values where |D2| dwarfs |D1| are ignored
        #print("criterion_met: ", criterion_met)

        # Use boolean indexing to select elements from delta1 where the criterion is True
        D1_selected = D1[criterion_met]
        D2_selected = D2[criterion_met]

        #print("**** delta1_selected: ", D1_selected)
        #print("*** delta2_selected: ", D2_selected)

        ratios = abs(D2_selected / D1_selected)
        #print(ratios)

        if len(ratios) == 0:
            return 0
        else:
            return np.max(ratios)    # An argument might be made for taking the avg SUM instead



    def norm_D(self, prev_conc: np.array, baseline_conc: np.array, delta_conc: np.array) -> float:
        """
        Return a measure of curvature in the concentration vs. time curves; larger values might be heralding
        onset of simulation instability

        :param prev_conc:       A numpy array with the concentration of the chemicals of interest,
                                    in the step prior to the current one (i.e. an "archive" value)
        :param baseline_conc:   A Numpy array with the concentration of the chemicals of interest
                                    at the start of a simulation time step
        :param delta_conc:      A Numpy array with the concentration changes
                                    of the chemicals of interest across a time step
        :return:
        """
        if prev_conc is None:
            return 0            # Unable to compute a norm; 0 represents "perfect"

        D1 = baseline_conc - prev_conc  # Change from prev state to current one
        D2 = delta_conc                 # Change from current state to next one

        #print("\nnorm_D ****** D1: ", D1)
        #print("norm_D ****** D2: ", D2)

        criterion_met = ~np.isclose(D1, 0) & (abs(D2) < 100 * abs(D1))
        # Note: values with D1 very close to zero are ignored
        #       likewise, values where |D2| dwarfs |D1| are ignored
        #print("criterion_met: ", criterion_met)

        # Use boolean indexing to select elements from delta1 where the criterion is True
        D1_selected = D1[criterion_met]
        D2_selected = D2[criterion_met]

        #print("norm_D **** delta1_selected: ", D1_selected)
        #print("norm_D *** delta2_selected: ", D2_selected)

        ratios = abs(D2_selected / D1_selected) # How big are the next changes relative to the previous ones
        #print(ratios)

        res = np.sum(ratios)    # An argument might be made for taking the MAX instead
        # TODO: turn into a separate norm
        arr_size = len(baseline_conc)
        normalized_res = res / arr_size
        #print("norm_D *** : ", normalized_res)

        return normalized_res



import math
import numpy as np
from typing import Tuple


class ThermoDynamics:
    """
    Manage the Thermodynamics aspects of reactions:
    changes in Gibbs Free Energy, Enthalpy, Entropy - and how
    they relate to equilibrium constant, at a given temperature.

    This class does NOT get instantiated.

            "K"       (equilibrium constant - from either kinetic or thermodynamic data;
                       if both present, they must match up!)
            "delta_H" (change in Enthalpy: Enthalpy of Products - Enthalpy of Reactants)
            "delta_S" (change in Entropy)
            "delta_G" (change in Gibbs Free Energy)

            Note - at constant temperature T :
                Delta_G = Delta_H - T * Delta_S
                Equilibrium constant = exp(-Delta_G / RT)
    """

    # Class Attribute
    R = 8.31446261815324    # Ideal gas constant, in units of Joules / (Kelvin * Mole)



    @classmethod
    def K_from_delta_G(cls, delta_G, temp) -> float:
        """
        Compute a reaction's equilibrium constant from the thermodynamic data

        :param delta_G: Change in Gibbs Free Energy (from reactants to products), in Joules
        :param temp:    System's temperature, in degree Kelvins
        :return:        The reaction's equilibrium constant
        """
        return math.exp(- delta_G / (cls.R * temp))



    @classmethod
    def delta_G_from_K(cls, K, temp) -> float:
        """
        Compute a reaction's change in its Gibbs Free Energy from its equilibrium constant,
        at the specified temperature

        :param K:       The reaction's equilibrium constant
        :param temp:    System's temperature, in degree Kelvins
        :return:        The reaction's change in Gibbs Free Energy (from reactants to products),
                            in Joules
        """
        return -cls.R * temp * math.log(K)      # Natural log



    @classmethod
    def delta_G_from_enthalpy(cls, delta_H, delta_S, temp) -> float:
        """
        Compute the change in Gibbs Free Energy, from Enthalpy and Entropy changes

        :param delta_H: The reaction's change in Enthalpy (from reactants to products)
        :param delta_S: The reaction's change in Entropy (from reactants to products)
        :param temp:    System's temperature, in degree Kelvins
        :return:        The reaction's change in Free Energy (from reactants to products)
        """
        return delta_H - temp * delta_S


    @classmethod
    def delta_H_from_gibbs(cls, delta_G, delta_S, temp) -> float:
        """
        Compute the change in Enthalpy, from changes in Gibbs Free Energy and in Entropy

        :param delta_G: The reaction's change in Gibbs Free Energy (from reactants to products)
        :param delta_S: The reaction's change in Entropy (from reactants to products)
        :param temp:    System's temperature, in degree Kelvins
        :return:        The reaction's change in Enthalpy (from reactants to products)
        """
        return delta_G + temp * delta_S


    @classmethod
    def delta_S_from_gibbs(cls, delta_G, delta_H, temp) -> float:
        """
        Compute the change in Entropy, from  changes in the Gibbs Free Energy and in Enthalpy

        :param delta_G: The reaction's change in Gibbs Free Energy (from reactants to products)
        :param delta_H: The reaction's change in Enthalpy (from reactants to products)
        :param temp:    System's temperature, in degree Kelvins
        :return:        The reaction's change in Entropy (from reactants to products)
        """
        return (delta_H - delta_G) / temp



    @classmethod
    def compute_reaction_quotient(cls, reactant_data :[str|Tuple[int, str]], product_data :[str|Tuple[int, str]],
                                  conc :dict, explain=False) -> np.double | Tuple[np.double, str]:
        """
        Compute the "Reaction Quotient" Q (aka "Mass–action Ratio"),
        for the reaction with the specified parameters,
        given the concentrations of chemicals involved in the reaction.

        EXAMPLE: use reactant_data=[(2, "A")] and product_data=[(1, "B")] ,
                 for a reaction of the form 2 A <-> B , 
                 alongside a dictionary with the concentrations (activities) of A and B

        Note: in a heterogeneous mixture, solids, pure liquids and solvents have an activity that has a fixed value of 1,
              and should be omitted from the parameters passed to this function.
              We're using the term "concentrations" instead of "chemical activities";
              concentrations approximate the activities of ideal dilute solutions

        :param reactant_data:   List of either STRINGS with the labels of the reactants,
                                    or PAIRS of the form (stoichiometry coefficient, label) of the reactants.
                                    If the stoichiometry coefficient isn't specified, it's taken to be 1
        :param product_data:    List of either STRINGS with the labels of the products of the reactions,
                                    or PAIRS of the form (stoichiometry coefficient, label) of the products.
                                    If the stoichiometry coefficient isn't specified, it's taken to be 1
        :param conc:            Dictionary with the concentrations (activities) of the species involved in the reaction.
                                The keys are the chemical labels
                                    EXAMPLE: {'A': 23.9, 'B': 36.1}
        :param explain:         If True, it also returns the math formula being used for the computation
                                    EXAMPLES:   "([C][D]) / ([A][B])"
                                                "[B] /  [A]^2 "

        :return:                If explain is False, return a value for the "Reaction Quotient" (aka "Mass–action Ratio");
                                    if True, return a pair with that quotient and a string with the math formula that was used.
                                    Note that the reaction quotient is a Numpy scalar that might be np.inf or np.nan
        """
        # TODO: could be tidier in avoiding unnecessary blanks in the explanations
        numerator = np.double(1)    # The product of all the concentrations of the reaction products (adjusted for reaction order)
        denominator = np.double(1)  # The product of all the concentrations of the reactants (also adjusted for reaction order)

        numerator_text = ""      # First part of the textual explanation
        denominator_text = ""    # Second part of the textual explanation


        # Compute the numerator of the "Reaction Quotient"
        for term in product_data:
            # Loop over the reaction products
            if type(term) == str:
                stoich_coeff = 1
                p = term
            else:
                (stoich_coeff, p) = term
                assert type(stoich_coeff) == int, f"compute_reaction_quotient(): the argument `product_data` " \
                                                  f"must be a list of pairs (integer and string).  `{stoich_coeff}` is not an integer"
                assert type(p) == str, f"compute_reaction_quotient(): the argument `product_data` " \
                                       f"must be a list of pairs (integer and string).  {p} is not a string"

            species_name = p
            # TODO: Maybe turn the several next lines into a helper function
            species_conc = conc.get(species_name)
            assert species_conc is not None, f"compute_reaction_quotient(): unable to proceed because the " \
                                             f"concentration of product `{species_name}` was not provided"

            numerator *= (species_conc ** stoich_coeff)
            if explain:
                if stoich_coeff > 1:
                    numerator_text += f" [{species_name}]^{stoich_coeff} "
                else:
                    numerator_text += f"[{species_name}]"

        if explain and len(product_data) > 1:
            numerator_text = f"({numerator_text})"  # In case of multiple terms, enclose them in parenthesis


        # Compute the denominator of the "Reaction Quotient"
        for term in reactant_data:
            # Loop over the reactants
            if type(term) == str:
                stoich_coeff = 1
                r = term
            else:
                (stoich_coeff, r) = term
                assert type(stoich_coeff) == int, f"compute_reaction_quotient(): the argument `reactant_data` " \
                                                  f"must be a list of pairs (integer and string).  `{stoich_coeff}` is not an integer"
                assert type(r) == str, f"compute_reaction_quotient(): the argument `reactant_data` " \
                                       f"must be a list of pairs (integer and string).  {r} is not a string"

            species_name =  r
            # TODO: Maybe turn the several next lines into a helper function
            species_conc = conc.get(species_name)
            assert species_conc is not None, f"compute_reaction_quotient(): unable to proceed because the " \
                                             f"concentration of reactant `{species_name}` was not provided"

            denominator *= (species_conc ** stoich_coeff)
            if explain:
                if stoich_coeff > 1:
                    denominator_text += f" [{species_name}]^{stoich_coeff} "
                else:
                    denominator_text += f"[{species_name}]"

        if explain and len(reactant_data) > 1:
            denominator_text = f"({denominator_text})"  # In case of multiple terms, enclose them in parenthesis


        with np.errstate(divide='ignore', invalid='ignore'):
            # It might be np.inf (if just the denominator is zero) or np.nan (if both are zero)
            quotient = numerator / denominator

        if explain:
            formula = f"{numerator_text} / {denominator_text}"
            return (quotient, formula)

        return quotient



    @classmethod
    def extract_thermodynamic_data(cls, temp, K=None, delta_H=None, delta_S=None, delta_G=None) -> dict:
        """

        :param temp:                System's temperature, in degree Kelvins
        :param K:       [OPTIONAL]  The reaction's equilibrium constant
        :param delta_H: [OPTIONAL]  The reaction's change in Enthalpy (from reactants to products)
        :param delta_S: [OPTIONAL]  The reaction's change in Entropy (from reactants to products)
        :param delta_G: [OPTIONAL]  The reaction's change in Gibbs Free Energy (from reactants to products)

        :return:        A dict with the 4 keys  "K", "delta_H", "delta_S", "delta_G"
        """
        #print(f"In extract_thermodynamic_data(): temp={temp}, K={K}, delta_H={delta_H}, delta_S={delta_S}, delta_G={delta_G}")

        assert temp is not None, \
            "extract_thermodynamic_data(): a temperature value (in K) must be passed to argument `temp`"


        if (K is not None):
            # If the temperature is set, compute the change in Gibbs Free Energy
            delta_G_derived = cls.delta_G_from_K(K = K, temp = temp)
            if delta_G is None:
                delta_G = delta_G_derived
            else:   # If already present (passed as argument), make sure that the two match!
                assert np.allclose(delta_G_derived, delta_G), \
                    f"extract_thermodynamic_data(): inconsistency between the derived ({delta_G_derived}) " \
                    f"and the passed ({delta_G}) values of Delta_G"


        if (delta_H is not None) and (delta_S is not None):
            # If all the thermodynamic data (possibly except delta_G) is available...

            # Compute the change in Gibbs Free Energy from delta_H and delta_S, at the current temperature
            delta_G_derived = cls.delta_G_from_enthalpy(delta_H = delta_H, delta_S = delta_S, temp = temp)

            if delta_G is None:
                delta_G = delta_G_derived
            else:  # If already present (passed as argument), make sure that the two match!
                assert np.allclose(delta_G_derived, delta_G), \
                    f"extract_thermodynamic_data(): inconsistency between the value of Delta_G ({delta_G_derived}) " \
                    f"derived from enthalpy/entropy, and the its value ({delta_G}) passed as argument or derived from K"


        if delta_G is not None:
            if K is None:
                # Compute the equilibrium constant (from the thermodynamic data)
                # Note: no need to do it if K is present, because we ALREADY checked for consistency
                K = ThermoDynamics.K_from_delta_G(delta_G = delta_G, temp = temp)

            # If either Enthalpy or Entropy is missing, but the other one is known, compute the missing one
            if (delta_H is None) and (delta_S is not None):
                delta_H = ThermoDynamics.delta_H_from_gibbs(delta_G=delta_G, delta_S=delta_S, temp=temp)
            elif (delta_H is not None) and (delta_S is None):
                delta_S = ThermoDynamics.delta_S_from_gibbs(delta_G=delta_G, delta_H=delta_H, temp=temp)


        return {"K": K, "delta_H": delta_H, "delta_S": delta_S, "delta_G": delta_G}
